"use strict";
/**
 * Created by rburson on 10/19/17.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("./util");
const moment = require("moment");
const dialog_1 = require("./dialog");
/*
 ************************** Dialog Models ****************************
 * These models correspond to those in the WebAPI schema specification
 * *******************************************************************
 */
/** ************** Base classes have to be defined first i.e. Order matters ********************/
class BinaryRef {
    constructor(_settings) {
        this._settings = _settings;
    }
    //@TODO
    /*
     static fromWSValue(encodedValue:string, settings:StringDictionary):Try<BinaryRef> {

     if (encodedValue && encodedValue.length > 0) {
     return new Success(new InlineBinaryRef(encodedValue, settings));
     } else {
     return new Success(new ObjectBinaryRef(settings));
     }

     }
     */
    get settings() {
        return this._settings;
    }
}
exports.BinaryRef = BinaryRef;
class CellValue {
    constructor(style) {
        this.style = style;
    }
    get isHeading1Style() {
        return this.style && (this.style === CellValue.STYLE_HEADING1);
    }
    get isHeading2Style() {
        return this.style && (this.style === CellValue.STYLE_HEADING2);
    }
    get isHeading3Style() {
        return this.style && (this.style === CellValue.STYLE_HEADING3);
    }
    get isHeading4Style() {
        return this.style && (this.style === CellValue.STYLE_HEADING4);
    }
    get isInlineMediaStyle() {
        return this.style && (this.style === CellValue.STYLE_INLINE_MEDIA || this.style === CellValue.STYLE_INLINE_MEDIA2);
    }
}
CellValue.STYLE_HEADING1 = "textHeading1";
CellValue.STYLE_HEADING2 = "textHeading2";
CellValue.STYLE_HEADING3 = "textHeading3";
CellValue.STYLE_HEADING4 = "textHeading4";
CellValue.STYLE_INLINE_MEDIA = "inlineMedia";
CellValue.STYLE_INLINE_MEDIA2 = "Image/Video";
exports.CellValue = CellValue;
/**
 * A View represents a Catavolt 'Pane' definition.  A Pane can be thought of as a 'panel' or UI component
 * that is responsible for displaying a data record or records. The Pane describes 'how' and 'where' the data will be
 * displayed, as well as surrounding 'meta' data (i.e. the Pane title, the Pane's menus).  The Pane itself does not contain
 * the record or records to be displayed, but may be combined with a {@link Record}(s) to display the data.
 */
class View {
    /* @TODO Leftover from PaneDef */
    /*
    readonly label:string;
    readonly viewDescs:Array<ViewDesc>;
    readonly recordDef:RecordDef;
    readonly dialogRedirection:DialogRedirection;
    readonly settings:StringDictionary;
    */
    /**
     * Find the title for this Pane
     * @returns {string}
     */
    findTitle() {
        var result = this.title ? this.title.trim() : '';
        result = result === 'null' ? '' : result;
        if (result === '') {
            //@TODO put this back when label is resolved
            //result = this.label ? this.label.trim() : '';
            //result = result === 'null' ? '' : result;
        }
        return result;
    }
    /**
     * Find a menu def on this View with the given actionId
     * @param actionId
     * @returns {Menu}
     */
    findMenuAt(actionId) {
        var result = null;
        if (this.menu) {
            this.menu.children.some((md) => {
                result = md.findAtActionId(actionId);
                return result != null;
            });
        }
        return result;
    }
}
exports.View = View;
/**
 * Defines how to present a business-value in a UI component
 */
class AttributeCellValue extends CellValue {
    constructor(propertyName, entryMethod, hint, tooltip, mask, autoFillCapable, actions, style) {
        super(style);
        this.propertyName = propertyName;
        this.entryMethod = entryMethod;
        this.hint = hint;
        this.tooltip = tooltip;
        this.mask = mask;
        this.autoFillCapable = autoFillCapable;
        this.actions = actions;
    }
    get isComboBoxEntryMethod() {
        return this.entryMethod && this.entryMethod === 'COMBO_BOX';
    }
    get isDropDownEntryMethod() {
        return this.entryMethod && this.entryMethod === 'DROP_DOWN';
    }
    get isIconEntryMethod() {
        return this.entryMethod && this.entryMethod === 'ICON_CHOOSER';
    }
    get isTextFieldEntryMethod() {
        return !this.entryMethod || this.entryMethod === 'TEXT_FIELD';
    }
}
exports.AttributeCellValue = AttributeCellValue;
/**
 * A purely declarative type. This object has no additional properties.
 */
class BarcodeScan extends View {
}
exports.BarcodeScan = BarcodeScan;
/**
 * Represents a base64 encoded binary
 */
class EncodedBinary {
    constructor(_data, _mimeType) {
        this._data = _data;
        this._mimeType = _mimeType;
    }
    /**
     * Get the base64 encoded data
     * @returns {string}
     */
    get data() {
        return this._data;
    }
    /**
     * Get the mime-type
     * @returns {string|string}
     */
    get mimeType() {
        return this._mimeType || 'application/octet-stream';
    }
    /**
     * Returns a 'data url' representation of this binary, including the encoded data
     * @returns {string}
     */
    toUrl() {
        return util_1.DataUrl.createDataUrl(this.mimeType, this.data);
    }
}
exports.EncodedBinary = EncodedBinary;
/**
 * Represents a remote binary
 */
class UrlBinary {
    constructor(_url) {
        this._url = _url;
    }
    get url() {
        return this._url;
    }
    /**
     * Returns a url that 'points to' the binary data
     * @returns {string}
     */
    toUrl() {
        return this.url;
    }
}
exports.UrlBinary = UrlBinary;
class Attachment {
    constructor(name, attachmentData) {
        this.name = name;
        this.attachmentData = attachmentData;
    }
    ;
}
exports.Attachment = Attachment;
/**
 * An abstract visual Calendar
 */
class Calendar extends View {
}
exports.Calendar = Calendar;
class CodeRef {
    constructor(code, description, type) {
        this.code = code;
        this.description = description;
        this.type = type;
    }
    toString() {
        return this.code + ":" + this.description;
    }
}
exports.CodeRef = CodeRef;
/**
 * A abstract definition for small visual area populated with labels and values. Labels are typically presented as simple text phrases.
 * Values are usually presented inside a UI component, such as a TextField or ComboBox.
 */
class Details extends View {
}
exports.Details = Details;
class ReferringObject {
    isDialogReferrer() {
        return this.type === TypeNames.ReferringDialogTypeName;
    }
    isWorkbenchReferrer() {
        return this.type === TypeNames.ReferringWorkbenchTypeName;
    }
}
exports.ReferringObject = ReferringObject;
class ReferringDialog extends ReferringObject {
}
exports.ReferringDialog = ReferringDialog;
class ReferringWorkbench extends ReferringObject {
}
exports.ReferringWorkbench = ReferringWorkbench;
class DialogException {
    constructor(iconName, message, name, stackTrace, title, cause, userMessages) {
        this.iconName = iconName;
        this.message = message;
        this.name = name;
        this.stackTrace = stackTrace;
        this.title = title;
        this.cause = cause;
        this.userMessages = userMessages;
    }
}
exports.DialogException = DialogException;
class DataAnnotation {
    constructor(name, value, type) {
        this.name = name;
        this.value = value;
        this.type = type;
    }
    static backgroundColor(annotations) {
        var result = util_1.ArrayUtil.find(annotations, (anno) => {
            return anno.isBackgroundColor;
        });
        return result ? result.backgroundColor : null;
    }
    static foregroundColor(annotations) {
        var result = util_1.ArrayUtil.find(annotations, (anno) => {
            return anno.isForegroundColor;
        });
        return result ? result.foregroundColor : null;
    }
    static imageName(annotations) {
        var result = util_1.ArrayUtil.find(annotations, (anno) => {
            return anno.isImageName;
        });
        return result ? result.value : null;
    }
    static imagePlacement(annotations) {
        var result = util_1.ArrayUtil.find(annotations, (anno) => {
            return anno.isImagePlacement;
        });
        return result ? result.value : null;
    }
    static isBoldText(annotations) {
        return annotations.some((anno) => {
            return anno.isBoldText;
        });
    }
    static isItalicText(annotations) {
        return annotations.some((anno) => {
            return anno.isItalicText;
        });
    }
    static isPlacementCenter(annotations) {
        return annotations.some((anno) => {
            return anno.isPlacementCenter;
        });
    }
    static isPlacementLeft(annotations) {
        return annotations.some((anno) => {
            return anno.isPlacementLeft;
        });
    }
    static isPlacementRight(annotations) {
        return annotations.some((anno) => {
            return anno.isPlacementRight;
        });
    }
    static isPlacementStretchUnder(annotations) {
        return annotations.some((anno) => {
            return anno.isPlacementStretchUnder;
        });
    }
    static isPlacementUnder(annotations) {
        return annotations.some((anno) => {
            return anno.isPlacementUnder;
        });
    }
    static isUnderlineText(annotations) {
        return annotations.some((anno) => {
            return anno.isUnderlineText;
        });
    }
    static overrideText(annotations) {
        var result = util_1.ArrayUtil.find(annotations, (anno) => {
            return anno.isOverrideText;
        });
        return result ? result.value : null;
    }
    static tipText(annotations) {
        var result = util_1.ArrayUtil.find(annotations, (anno) => {
            return anno.isTipText;
        });
        return result ? result.value : null;
    }
    get backgroundColor() {
        return this.isBackgroundColor ? this.value : null;
    }
    get foregroundColor() {
        return this.isForegroundColor ? this.value : null;
    }
    equals(dataAnno) {
        return this.name === dataAnno.name;
    }
    get isBackgroundColor() {
        return this.name === DataAnnotation.BACKGROUND_COLOR;
    }
    get isBoldText() {
        return this.name === DataAnnotation.BOLD_TEXT && this.value === DataAnnotation.TRUE_VALUE;
    }
    get isForegroundColor() {
        return this.name === DataAnnotation.FOREGROUND_COLOR;
    }
    get isImageName() {
        return this.name === DataAnnotation.IMAGE_NAME;
    }
    get isImagePlacement() {
        return this.name === DataAnnotation.IMAGE_PLACEMENT;
    }
    get isItalicText() {
        return this.name === DataAnnotation.ITALIC_TEXT && this.value === DataAnnotation.TRUE_VALUE;
    }
    get isOverrideText() {
        return this.name === DataAnnotation.OVERRIDE_TEXT;
    }
    get isPlacementCenter() {
        return this.isImagePlacement && this.value === DataAnnotation.PLACEMENT_CENTER;
    }
    get isPlacementLeft() {
        return this.isImagePlacement && this.value === DataAnnotation.PLACEMENT_LEFT;
    }
    get isPlacementRight() {
        return this.isImagePlacement && this.value === DataAnnotation.PLACEMENT_RIGHT;
    }
    get isPlacementStretchUnder() {
        return this.isImagePlacement && this.value === DataAnnotation.PLACEMENT_STRETCH_UNDER;
    }
    get isPlacementUnder() {
        return this.isImagePlacement && this.value === DataAnnotation.PLACEMENT_UNDER;
    }
    get isTipText() {
        return this.name === DataAnnotation.TIP_TEXT;
    }
    get isUnderlineText() {
        return this.name === DataAnnotation.UNDERLINE && this.value === DataAnnotation.TRUE_VALUE;
    }
    toJSON() {
        return { name: this.name, value: this.value, type: this.type };
    }
}
DataAnnotation.BOLD_TEXT = "BOLD_TEXT";
DataAnnotation.BACKGROUND_COLOR = "BGND_COLOR";
DataAnnotation.FOREGROUND_COLOR = "FGND_COLOR";
DataAnnotation.IMAGE_NAME = "IMAGE_NAME";
DataAnnotation.IMAGE_PLACEMENT = "IMAGE_PLACEMENT";
DataAnnotation.ITALIC_TEXT = "ITALIC_TEXT";
DataAnnotation.OVERRIDE_TEXT = "OVRD_TEXT";
DataAnnotation.TIP_TEXT = "TIP_TEXT";
DataAnnotation.UNDERLINE = "UNDERLINE";
DataAnnotation.TRUE_VALUE = "1";
DataAnnotation.PLACEMENT_CENTER = "CENTER";
DataAnnotation.PLACEMENT_LEFT = "LEFT";
DataAnnotation.PLACEMENT_RIGHT = "RIGHT";
DataAnnotation.PLACEMENT_UNDER = "UNDER";
DataAnnotation.PLACEMENT_STRETCH_UNDER = "STRETCH_UNDER";
exports.DataAnnotation = DataAnnotation;
/**
 * Utility for working with Records
 */
class RecordUtil {
    static newRecord(id, properties, annotations, type) {
        return annotations ? new RecordImpl(id, util_1.ArrayUtil.copy(properties), util_1.ArrayUtil.copy(annotations), type) :
            new RecordImpl(id, util_1.ArrayUtil.copy(properties), null, type);
    }
    static isRecord(o) {
        return (o instanceof RecordImpl)
            || (o instanceof RecordBuffer)
            || (o instanceof NullRecord);
    }
}
exports.RecordUtil = RecordUtil;
/**
 * An {@link Record} that manages two copies internally, a before and after, for 'undo' and comparison purposes.
 * An Record Represents a 'Record' or set of {@link Prop} (names and values).
 * An Record may also have {@link Annotation}s (style annotations) that apply to the whole 'record'
 */
class RecordBuffer {
    constructor(_before, _after) {
        this._before = _before;
        this._after = _after;
        if (!_before)
            throw new Error('_before is null in RecordBuffer');
        if (!_after)
            this._after = _before;
    }
    static createRecordBuffer(id, before, after, annotations, type) {
        return new RecordBuffer(RecordUtil.newRecord(id, before, annotations, type), RecordUtil.newRecord(id, after, annotations, type));
    }
    get after() {
        return this._after;
    }
    get annotations() {
        return this._after.annotations;
    }
    annotationsAtName(propName) {
        return this._after.annotationsAtName(propName);
    }
    afterEffects(afterAnother) {
        if (afterAnother) {
            return this._after.afterEffects(afterAnother);
        }
        else {
            return this._before.afterEffects(this._after);
        }
    }
    get backgroundColor() {
        return this._after.backgroundColor;
    }
    backgroundColorFor(propName) {
        return this._after.backgroundColorFor(propName);
    }
    get before() {
        return this._before;
    }
    get foregroundColor() {
        return this._after.foregroundColor;
    }
    foregroundColorFor(propName) {
        return this._after.foregroundColorFor(propName);
    }
    get imageName() {
        return this._after.imageName;
    }
    imageNameFor(propName) {
        return this._after.imageNameFor(propName);
    }
    get imagePlacement() {
        return this._after.imagePlacement;
    }
    imagePlacementFor(propName) {
        return this._after.imagePlacement;
    }
    get isBoldText() {
        return this._after.isBoldText;
    }
    isBoldTextFor(propName) {
        return this._after.isBoldTextFor(propName);
    }
    isChanged(name) {
        var before = this._before.propAtName(name);
        var after = this._after.propAtName(name);
        return (before && after) ? !before.equals(after) : !(!before && !after);
    }
    get isItalicText() {
        return this._after.isItalicText;
    }
    isItalicTextFor(propName) {
        return this._after.isItalicTextFor(propName);
    }
    get isPlacementCenter() {
        return this._after.isPlacementCenter;
    }
    isPlacementCenterFor(propName) {
        return this._after.isPlacementCenterFor(propName);
    }
    get isPlacementLeft() {
        return this._after.isPlacementLeft;
    }
    isPlacementLeftFor(propName) {
        return this._after.isPlacementLeftFor(propName);
    }
    get isPlacementRight() {
        return this._after.isPlacementRight;
    }
    isPlacementRightFor(propName) {
        return this._after.isPlacementRightFor(propName);
    }
    get isPlacementStretchUnder() {
        return this._after.isPlacementStretchUnder;
    }
    isPlacementStretchUnderFor(propName) {
        return this._after.isPlacementStretchUnderFor(propName);
    }
    get isPlacementUnder() {
        return this._after.isPlacementUnder;
    }
    isPlacementUnderFor(propName) {
        return this._after.isPlacementUnderFor(propName);
    }
    get isUnderline() {
        return this._after.isUnderline;
    }
    isUnderlineFor(propName) {
        return this._after.isUnderlineFor(propName);
    }
    get id() {
        return this._after.id;
    }
    get overrideText() {
        return this._after.overrideText;
    }
    overrideTextFor(propName) {
        return this._after.overrideTextFor(propName);
    }
    propAtIndex(index) {
        return this.properties[index];
    }
    propAtName(propName) {
        return this._after.propAtName(propName);
    }
    get propCount() {
        return this._after.propCount;
    }
    get propNames() {
        return this._after.propNames;
    }
    get properties() {
        return this._after.properties;
    }
    get propValues() {
        return this._after.propValues;
    }
    get type() {
        return this._after.type;
    }
    setValue(name, value) {
        const newProps = [];
        let found = false;
        this.properties.forEach((prop) => {
            if (prop.name === name) {
                newProps.push(new Property(name, value, prop.propertyType, prop.format, prop.annotations));
                found = true;
            }
            else {
                newProps.push(prop);
            }
        });
        if (!found) {
            newProps.push(new Property(name, value));
        }
        this._after = RecordUtil.newRecord(this.id, newProps, this.annotations, this.type);
    }
    get tipText() {
        return this._after.tipText;
    }
    tipTextFor(propName) {
        return this._after.tipTextFor(propName);
    }
    toRecord() {
        return RecordUtil.newRecord(this.id, this.properties, this.annotations, this.type);
    }
    toJSON() {
        return this.afterEffects().toJSON();
    }
    valueAtName(propName) {
        return this._after.valueAtName(propName);
    }
}
exports.RecordBuffer = RecordBuffer;
/**
 * *********************************
 */
/**
 * The implementation of {@link Record}.
 * Represents a 'Record' or set of {@link Prop} (names and values).
 * An Record may also have {@link Annotation}s (style annotations) that apply to the whole 'record'
 */
class RecordImpl {
    constructor(id, properties = [], annotations = [], type) {
        this.id = id;
        this.properties = properties;
        this.annotations = annotations;
        this.type = type;
    }
    annotationsAtName(propName) {
        var p = this.propAtName(propName);
        return p ? p.annotations : [];
    }
    afterEffects(after) {
        var effects = [];
        after.properties.forEach((afterProp) => {
            var beforeProp = this.propAtName(afterProp.name);
            if (!afterProp.equals(beforeProp)) {
                effects.push(afterProp);
            }
        });
        return new RecordImpl(after.id, effects, after.annotations, after.type);
    }
    get backgroundColor() {
        return DataAnnotation.backgroundColor(this.annotations);
    }
    backgroundColorFor(propName) {
        var p = this.propAtName(propName);
        return p && p.backgroundColor ? p.backgroundColor : this.backgroundColor;
    }
    get foregroundColor() {
        return DataAnnotation.foregroundColor(this.annotations);
    }
    foregroundColorFor(propName) {
        var p = this.propAtName(propName);
        return p && p.foregroundColor ? p.foregroundColor : this.foregroundColor;
    }
    get imageName() {
        return DataAnnotation.imageName(this.annotations);
    }
    imageNameFor(propName) {
        var p = this.propAtName(propName);
        return p && p.imageName ? p.imageName : this.imageName;
    }
    get imagePlacement() {
        return DataAnnotation.imagePlacement(this.annotations);
    }
    imagePlacementFor(propName) {
        var p = this.propAtName(propName);
        return p && p.imagePlacement ? p.imagePlacement : this.imagePlacement;
    }
    get isBoldText() {
        return DataAnnotation.isBoldText(this.annotations);
    }
    isBoldTextFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isBoldText ? p.isBoldText : this.isBoldText;
    }
    get isItalicText() {
        return DataAnnotation.isItalicText(this.annotations);
    }
    isItalicTextFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isItalicText ? p.isItalicText : this.isItalicText;
    }
    get isPlacementCenter() {
        return DataAnnotation.isPlacementCenter(this.annotations);
    }
    isPlacementCenterFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isPlacementCenter ? p.isPlacementCenter : this.isPlacementCenter;
    }
    get isPlacementLeft() {
        return DataAnnotation.isPlacementLeft(this.annotations);
    }
    isPlacementLeftFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isPlacementLeft ? p.isPlacementLeft : this.isPlacementLeft;
    }
    get isPlacementRight() {
        return DataAnnotation.isPlacementRight(this.annotations);
    }
    isPlacementRightFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isPlacementRight ? p.isPlacementRight : this.isPlacementRight;
    }
    get isPlacementStretchUnder() {
        return DataAnnotation.isPlacementStretchUnder(this.annotations);
    }
    isPlacementStretchUnderFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isPlacementStretchUnder ? p.isPlacementStretchUnder : this.isPlacementStretchUnder;
    }
    get isPlacementUnder() {
        return DataAnnotation.isPlacementUnder(this.annotations);
    }
    isPlacementUnderFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isPlacementUnder ? p.isPlacementUnder : this.isPlacementUnder;
    }
    get isUnderline() {
        return DataAnnotation.isUnderlineText(this.annotations);
    }
    isUnderlineFor(propName) {
        var p = this.propAtName(propName);
        return p && p.isUnderline ? p.isUnderline : this.isUnderline;
    }
    get overrideText() {
        return DataAnnotation.overrideText(this.annotations);
    }
    overrideTextFor(propName) {
        var p = this.propAtName(propName);
        return p && p.overrideText ? p.overrideText : this.overrideText;
    }
    propAtIndex(index) {
        return this.properties[index];
    }
    propAtName(propName) {
        var prop = null;
        this.properties.some((p) => {
            if (p.name === propName) {
                prop = p;
                return true;
            }
            return false;
        });
        return prop;
    }
    get propCount() {
        return this.properties.length;
    }
    get propNames() {
        return this.properties.map((p) => {
            return p.name;
        });
    }
    get propValues() {
        return this.properties.map((p) => {
            return p.value;
        });
    }
    get tipText() {
        return DataAnnotation.tipText(this.annotations);
    }
    tipTextFor(propName) {
        var p = this.propAtName(propName);
        return p && p.tipText ? p.tipText : this.tipText;
    }
    toJSON() {
        return {
            id: this.id,
            properties: this.properties,
            type: TypeNames.RecordTypeName
        };
    }
    toRecord() {
        return this;
    }
    valueAtName(propName) {
        var value = null;
        this.properties.some((p) => {
            if (p.name === propName) {
                value = p.value;
                return true;
            }
            return false;
        });
        return value;
    }
}
exports.RecordImpl = RecordImpl;
/**
 * *********************************
 */
class ErrorMessage extends View {
}
exports.ErrorMessage = ErrorMessage;
/**
 * A purely declarative type. This object has no additional properties.
 */
class ForcedLineCellValue extends CellValue {
    constructor(style) {
        super(style);
    }
}
exports.ForcedLineCellValue = ForcedLineCellValue;
/**
 * A composition of View objects that, together, comprise a UI form.
 */
class Form extends View {
    get isCompositeForm() {
        return this.formStyle === 'COMPOSITE_FORM';
    }
    get isFlowingLayout() {
        return this.formLayout && this.formLayout === 'FLOWING';
    }
    get isFlowingTopDownLayout() {
        return this.formLayout && this.formLayout === 'FLOWING_TOP_DOWN';
    }
    get isFourBoxSquareLayout() {
        return (this.formLayout && this.formLayout === 'FOUR_BOX_SQUARE') ||
            (this.formLayout && this.formLayout === 'H(2,2)');
    }
    get isHorizontalLayout() {
        return this.formLayout && this.formLayout === 'H';
    }
    get isOptionsFormLayout() {
        return this.formLayout && this.formLayout === 'OPTIONS_FORM';
    }
    get isTabsLayout() {
        return this.formLayout && this.formLayout === 'TABS';
    }
    get isThreeBoxOneLeftLayout() {
        return this.formLayout && this.formLayout === 'THREE_ONE_LEFT';
    }
    get isThreeBoxOneOverLayout() {
        return this.formLayout && this.formLayout === 'THREE_ONE_OVER';
    }
    get isThreeBoxOneRightLayout() {
        return this.formLayout && this.formLayout === 'THREE_ONE_RIGHT';
    }
    get isThreeBoxOneUnderLayout() {
        return this.formLayout && this.formLayout === 'THREE_ONE_UNDER';
    }
    get isTopDownLayout() {
        return this.formLayout && this.formLayout === 'TOP_DOWN';
    }
    get isTwoVerticalLayout() {
        return this.formLayout && this.formLayout.indexOf('H(2,V') > -1;
    }
}
exports.Form = Form;
/**
 * A purely declarative type. This object has no additional properties.
 */
class GpsReading extends View {
}
exports.GpsReading = GpsReading;
class GpsReadingProperty {
    constructor(accuracy, latitude, longitude, source, type) {
        this.accuracy = accuracy;
        this.latitude = latitude;
        this.longitude = longitude;
        this.source = source;
        this.type = type;
    }
}
exports.GpsReadingProperty = GpsReadingProperty;
/**
 * A purely declarative type. This object has no additional properties.
 */
class MapLocation extends View {
}
exports.MapLocation = MapLocation;
class MapLocationProperty {
    constructor(latitude, longitude, type) {
        this.latitude = latitude;
        this.longitude = longitude;
        this.type = type;
    }
    ;
}
exports.MapLocationProperty = MapLocationProperty;
/**
 * A view describing how to display a collection of data as a line graph, pie chart, bar chart, etc.
 */
class Graph extends View {
}
Graph.GRAPH_TYPE_CARTESIAN = "GRAPH_TYPE_BAR";
Graph.GRAPH_TYPE_PIE = "GRAPH_TYPE_PIE";
Graph.PLOT_TYPE_BAR = "BAR";
Graph.PLOT_TYPE_BUBBLE = "BUBBLE";
Graph.PLOT_TYPE_LINE = "LINE";
Graph.PLOT_TYPE_SCATTER = "SCATTER";
Graph.PLOT_TYPE_STACKED = "STACKED";
exports.Graph = Graph;
class ImagePicker extends View {
}
exports.ImagePicker = ImagePicker;
class InlineBinaryRef extends BinaryRef {
    constructor(_inlineData, settings) {
        super(settings);
        this._inlineData = _inlineData;
    }
    /* Base64 encoded data */
    get inlineData() {
        return this._inlineData;
    }
    toString() {
        return this._inlineData;
    }
}
exports.InlineBinaryRef = InlineBinaryRef;
/**
 * A text description typically preceeding a UI component as a prompt
 */
class LabelCellValue extends CellValue {
    constructor(style, value) {
        super(style);
        this.value = value;
    }
}
exports.LabelCellValue = LabelCellValue;
/**
 * Columns, filter and sorts for a UI list component.
 */
class List extends View {
    get isDefaultStyle() {
        return this.style && this.style === 'DEFAULT';
    }
    get isDetailsFormStyle() {
        return this.style && this.style === 'DETAILS_FORM';
    }
    get isFormStyle() {
        return this.style && this.style === 'FORM';
    }
    get isTabularStyle() {
        return this.style && this.style === 'TABULAR';
    }
    get columnHeadings() {
        return this.columns.map((c) => {
            return c.heading;
        });
    }
    rowValues(record) {
        return this.columns.map((c) => {
            return record.valueAtName(c.propertyName);
        });
    }
}
exports.List = List;
class Map extends View {
}
exports.Map = Map;
class Menu {
    constructor() {
        this.children = [];
    }
    static findSubMenu(md, matcher) {
        if (matcher(md))
            return md;
        if (md.children) {
            for (let i = 0; i < md.children.length; i++) {
                let result = Menu.findSubMenu(md.children[i], matcher);
                if (result)
                    return result;
            }
        }
        return null;
    }
    findAtActionId(actionId) {
        if (this.actionId === actionId)
            return this;
        var result = null;
        if (this.children) {
            this.children.some((md) => {
                result = md.findAtActionId(actionId);
                return result != null;
            });
        }
        return result;
    }
    findContextMenu() {
        return Menu.findSubMenu(this, (md) => {
            return md.id === 'CONTEXT_MENU';
        });
    }
    get isPresaveDirective() {
        return this.directive && this.directive === 'PRESAVE';
    }
    get isRead() {
        return this.modes && this.modes.indexOf('R') > -1;
    }
    get isSeparator() {
        return this.type && this.type === 'separator';
    }
    get isWrite() {
        return this.modes && this.modes.indexOf('W') > -1;
    }
}
exports.Menu = Menu;
/**
 * An empty or uninitialized {@link Record}.
 * Represents a 'Record' or set of {@link Prop} (names and values).
 * An Record may also have {@link Annotation}s (style annotations) that apply to the whole 'record'
 */
class NullRecord {
    constructor() {
    }
    get annotations() {
        return [];
    }
    annotationsAtName(propName) {
        return [];
    }
    afterEffects(after) {
        return after;
    }
    get backgroundColor() {
        return null;
    }
    backgroundColorFor(propName) {
        return null;
    }
    get foregroundColor() {
        return null;
    }
    foregroundColorFor(propName) {
        return null;
    }
    get id() {
        return null;
    }
    get imageName() {
        return null;
    }
    imageNameFor(propName) {
        return null;
    }
    get imagePlacement() {
        return null;
    }
    imagePlacementFor(propName) {
        return null;
    }
    get isBoldText() {
        return false;
    }
    isBoldTextFor(propName) {
        return false;
    }
    get isItalicText() {
        return false;
    }
    isItalicTextFor(propName) {
        return false;
    }
    get isPlacementCenter() {
        return false;
    }
    isPlacementCenterFor(propName) {
        return false;
    }
    get isPlacementLeft() {
        return false;
    }
    isPlacementLeftFor(propName) {
        return false;
    }
    get isPlacementRight() {
        return false;
    }
    isPlacementRightFor(propName) {
        return false;
    }
    get isPlacementStretchUnder() {
        return false;
    }
    isPlacementStretchUnderFor(propName) {
        return false;
    }
    get isPlacementUnder() {
        return false;
    }
    isPlacementUnderFor(propName) {
        return false;
    }
    get isUnderline() {
        return false;
    }
    isUnderlineFor(propName) {
        return false;
    }
    get objectId() {
        return null;
    }
    get overrideText() {
        return null;
    }
    overrideTextFor(propName) {
        return null;
    }
    propAtIndex(index) {
        return null;
    }
    propAtName(propName) {
        return null;
    }
    get propCount() {
        return 0;
    }
    get propNames() {
        return [];
    }
    get properties() {
        return [];
    }
    get propValues() {
        return [];
    }
    get tipText() {
        return null;
    }
    get type() {
        return null;
    }
    tipTextFor(propName) {
        return null;
    }
    toRecord() {
        return this;
    }
    toJSON() {
        return {
            id: this.id,
            properties: this.properties,
            type: TypeNames.RecordTypeName
        };
    }
    valueAtName(propName) {
        return null;
    }
}
NullRecord.singleton = new NullRecord();
exports.NullRecord = NullRecord;
class ObjectBinaryRef extends BinaryRef {
    constructor(settings) {
        super(settings);
    }
}
exports.ObjectBinaryRef = ObjectBinaryRef;
class ObjectRef {
    constructor(objectId, description, type) {
        this.objectId = objectId;
        this.description = description;
        this.type = type;
    }
    toString() {
        return this.objectId + ":" + this.description;
    }
}
exports.ObjectRef = ObjectRef;
class PrintMarkup extends View {
    constructor() {
        super(...arguments);
        this._orderedCellValue = null;
    }
    get orderedCellValues() {
        if (!this._orderedCellValue) {
            this._orderedCellValue = {};
            this.rows.forEach((cellRow, index) => {
                cellRow.forEach((cell) => {
                    cell.values.forEach((cellValue) => {
                        if (cellValue instanceof AttributeCellValue) {
                            let attributeCellValue = cellValue;
                            this._orderedCellValue[attributeCellValue.propertyName] = attributeCellValue;
                        }
                    });
                });
            });
        }
        ;
        return this._orderedCellValue;
    }
}
exports.PrintMarkup = PrintMarkup;
/**
 * Represents a 'value' or field in a row or record. See {@link Record}
 * A Prop has a corresponding {@link PropDef} that describes the property.
 * Like an {@link Record}, a Prop may also have {@link Annotation}s (style annotations),
 * but these apply to the property only
 */
class Property {
    /**
     *
     * @param {string} name
     * @param value
     * @param {string} propertyType
     * @param {string} format
     * @param {Array<DataAnnotation>} annotations
     */
    constructor(name, value, propertyType, format, annotations = []) {
        this.name = name;
        this.value = value;
        this.propertyType = propertyType;
        this.format = format;
        this.annotations = annotations;
    }
    /**
     * Produce an unique string that can be used for comparison purposes
     * Props considered 'equal' should produce the same identity string
     *
     * @param o
     * @param propDef
     * @returns {any}
     */
    static identity(o, propDef) {
        if (typeof o === 'number') {
            return String(o);
        }
        else if (typeof o === 'object') {
            if (o instanceof Date) {
                return String(o.getTime());
            }
            else if (o instanceof util_1.DateValue) {
                return String(o.dateObj.getTime());
            }
            else if (o instanceof util_1.DateTimeValue) {
                return String(o.dateObj.getTime());
            }
            else if (o instanceof util_1.TimeValue) {
                return o.toString();
            }
            else if (o instanceof CodeRef) {
                return o.code;
            }
            else if (o instanceof ObjectRef) {
                return o.objectId;
            }
            else if (o instanceof GpsReading) {
                return o.toString();
            }
            else if (o instanceof MapLocation) {
                return o.toString();
            }
            else {
                return String(o);
            }
        }
        else {
            return String(o);
        }
    }
    static fromJSON(jsonObject) {
        return new Property(jsonObject.name, Property.parseValue(jsonObject.value, jsonObject.format), jsonObject.propertyType, jsonObject.format, jsonObject.annotations);
    }
    equals(prop) {
        return this.name === prop.name && this.value === prop.value;
    }
    get backgroundColor() {
        return DataAnnotation.backgroundColor(this.annotations);
    }
    get foregroundColor() {
        return DataAnnotation.foregroundColor(this.annotations);
    }
    get imageName() {
        return DataAnnotation.imageName(this.annotations);
    }
    get imagePlacement() {
        return DataAnnotation.imagePlacement(this.annotations);
    }
    get isBoldText() {
        return DataAnnotation.isBoldText(this.annotations);
    }
    get isItalicText() {
        return DataAnnotation.isItalicText(this.annotations);
    }
    get isPlacementCenter() {
        return DataAnnotation.isPlacementCenter(this.annotations);
    }
    get isPlacementLeft() {
        return DataAnnotation.isPlacementLeft(this.annotations);
    }
    get isPlacementRight() {
        return DataAnnotation.isPlacementRight(this.annotations);
    }
    get isPlacementStretchUnder() {
        return DataAnnotation.isPlacementStretchUnder(this.annotations);
    }
    get isPlacementUnder() {
        return DataAnnotation.isPlacementUnder(this.annotations);
    }
    get isUnderline() {
        return DataAnnotation.isUnderlineText(this.annotations);
    }
    get overrideText() {
        return DataAnnotation.overrideText(this.annotations);
    }
    get tipText() {
        return DataAnnotation.tipText(this.annotations);
    }
    get valueForWrite() {
        const o = this.value;
        if (typeof o === 'number') {
            return String(o);
        }
        else if (typeof o === 'object') {
            if (o instanceof Date) {
                //remove the 'Z' from the end of the ISO string for now, until the server supports timezones...
                return o.toISOString().slice(0, -1);
            }
            else if (o instanceof util_1.DateTimeValue) {
                //remove the 'Z' from the end of the ISO string for now, until the server supports timezones...
                return o.dateObj.toISOString().slice(0, -1);
            }
            else if (o instanceof util_1.DateValue) {
                //remove all Time information from the end of the ISO string from the 'T' to the end...
                const isoString = o.dateObj.toISOString();
                return isoString.slice(0, isoString.indexOf('T'));
            }
            else if (o instanceof util_1.TimeValue) {
                return o.toString();
            }
            else {
                //for any other type of value, return the object itself
                //this could include string, Array, CodeRef, ObjectRef, GpsReadingProperty, MapLocationProperty, InlineBinaryRef
                return o;
            }
        }
        else {
            return o;
        }
    }
    toJSON() {
        const jsonObject = {
            name: this.name,
            value: this.valueForWrite,
            type: TypeNames.PropertyTypeName
        };
        if (this.propertyType)
            jsonObject['propertyType'] = this.propertyType;
        if (this.format)
            jsonObject['format'] = this.format;
        return jsonObject;
    }
    static parseValue(value, format) {
        if (typeof value === 'string' && format) {
            if (['integer', 'decimal', 'int32', 'int64', 'float', 'double'].some(v => format === v)) {
                return Number(value);
            }
            else if (format === 'date') {
                //parse as ISO - no offset specified by server right now, so we assume local time
                return moment(value, 'YYYY-M-D').toDate();
            }
            else if (format === 'date-time') {
                //parse as ISO - no offset specified by server right now, so we assume local time
                //strip invalid suffix (sometimes) provided by server
                const i = value.indexOf('T0:');
                return moment((i > -1) ? value.substring(0, i) : value).toDate();
            }
            else if (format === 'time') {
                util_1.TimeValue.fromString(value);
            }
            else {
                return value;
            }
        }
        else {
            return value;
        }
    }
}
exports.Property = Property;
/**
 /**
 * A property definition describes a particular value of a business entity. Business entities are transacted as records,
 * therefore properties and lists of properties are referred to as fields and records. Moreover, a list of property definitions
 * is referred to as a record definition and is the metadata describing the read/write capabilities of a specific dialog model
 * in use by a specific user in a specific workflow.
 * Contains information that 'defines' a property {@link Prop} (name/value)
 * An instance of the {@link Property} contains the actual data value.
 */
class PropertyDef {
    constructor(/**
                     * The canCauseSideEffects meta property indicates that writing to this property can cause LOCAL side effects
                     * (on the same business object). For example, changing a 'zipCode' property case cause the 'state' property to change. If a user interface changes a property that can cause side effects, it should refresh the associated business view.
                     */ canCauseSideEffects, contentType, 
        /**
         * Length of a type to be displayed. Some types are longer than what is practically needed by the application.
         * This property is used to define the practical length used in user interfaces.
         */
        displayLength, 
        /**
         * Scale of a decimal type to be displayed. Some decimal types are longer than what is practically needed by
         * the application. This property is used to define the practical scale used in user interfaces.
         */
        displayScale, 
        /**
         * The format property further describes the value using names that correlate, when possible, to the
         * Open API formats. Some example format names are date, date-time, uuid, int32 and int64. The format name decimal
         * is used to describe a string holding an arbitrary precision BCD value. For more information,
         * see the Open API Spec at https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md
         */
        format, 
        /**
         * Length of a type. This can be the length of a string or the length of a decimal.
         */
        length, 
        /**
         * The name of a business-data value
         */
        propertyName, 
        /**
         * Scale of a decimal type
         */
        scale, 
        /**
         * Whereas 'type' and 'format' define the physical meaning of a property, the 'semanticType' adds meaningful
         * insight into the usage of the property. For example, a 'decimal' type may be further defined semantically as a 'money' type.
         */
        semanticType, 
        /**
         * The type of business-data value. A valid type is either one of the OpenAPI basic values
         * [array boolean integer null number object string] or a fully qualified type name that can be retrieved from the schema service.
         */
        propertyType, writeAllowed, writeEnabled) {
        this.canCauseSideEffects = canCauseSideEffects;
        this.contentType = contentType;
        this.displayLength = displayLength;
        this.displayScale = displayScale;
        this.format = format;
        this.length = length;
        this.propertyName = propertyName;
        this.scale = scale;
        this.semanticType = semanticType;
        this.propertyType = propertyType;
        this.writeAllowed = writeAllowed;
        this.writeEnabled = writeEnabled;
    }
    get isBarcodeType() {
        return this.semanticType === 'BARCODE';
    }
    get isBinaryType() {
        return this.isLargeBinaryType || this.isSignatureType;
    }
    get isBooleanType() {
        return this.propertyType === 'boolean';
    }
    get isCodeRefType() {
        return this.propertyType === TypeNames.CodeRefTypeName;
    }
    get isDateType() {
        return this.format === 'date';
    }
    get isDateTimeType() {
        return this.format === 'date-time';
    }
    get isDecimalType() {
        return this.format === 'decimal';
    }
    get isDoubleType() {
        return this.format === 'double';
    }
    get isEmailType() {
        return this.semanticType === 'EMAIL';
    }
    get isFileAttachment() {
        return this.semanticType === 'FILE_UPLOAD';
    }
    get isFloatType() {
        return this.format === 'float';
    }
    get isGpsReadingType() {
        return this.propertyType === TypeNames.GpsReadingPropertyTypeName;
    }
    get isMapLocationType() {
        return this.propertyType === TypeNames.MapLocationPropertyTypeName;
    }
    get isHTMLType() {
        return this.semanticType === 'DATA_HTML';
    }
    //@TODO
    get isInlineMediaStyle() {
        return (this.semanticType === CellValue.STYLE_INLINE_MEDIA || this.semanticType === CellValue.STYLE_INLINE_MEDIA2);
    }
    get isListType() {
        return this.propertyType === 'array';
    }
    get isIntType() {
        return ['integer', 'int32', 'int64'].some(v => this.propertyType === v);
    }
    get isLargeBinaryType() {
        return this.semanticType === 'LARGE_BINARY';
    }
    get isLongType() {
        return this.format === 'int64';
    }
    get isMoneyType() {
        return this.semanticType === 'MONEY';
    }
    get isNumericType() {
        return this.isDecimalType || this.isIntType || this.isDoubleType || this.isLongType || this.isFloatType;
    }
    get isObjRefType() {
        return this.propertyType === TypeNames.ObjectRefTypeName;
    }
    get isPasswordType() {
        return this.format === 'password' || this.semanticType === 'PASSWORD';
    }
    get isPercentType() {
        return this.semanticType === 'PERCENT';
    }
    get isSignatureType() {
        return this.semanticType === 'USER_SIGNATURE';
    }
    get isStringType() {
        return this.propertyType === 'string';
    }
    get isTelephoneType() {
        return this.semanticType === 'TELEPHONE';
    }
    get isTextBlock() {
        return this.semanticType === 'TEXT_BLOCK';
    }
    get isTimeType() {
        return this.format === 'time';
    }
    get isUnformattedNumericType() {
        return this.semanticType === 'UNFORMATTED';
    }
    get isURLType() {
        return this.semanticType === 'URL';
    }
}
exports.PropertyDef = PropertyDef;
/**
 * ************* Property Formatting ********************
 */
/**
 * Helper for transforming values to and from formats suitable for reading and writing to the server
 * (i.e. object to string and string to object)
 */
class PrivatePropFormats {
}
PrivatePropFormats.decimalFormat = ["0,0", "0,0.0", "0,0.00", "0,0.000", "0,0.0000", "0,0.00000", "0,0.000000", "0,0.0000000", "0,0.00000000", "0,0.000000000", "0,0.0000000000"];
PrivatePropFormats.decimalFormatGeneric = "0,0.[0000000000000000000000000]";
PrivatePropFormats.moneyFormat = ["$0,0", "$0,0.0", "$0,0.00", "$0,0.000", "$0,0.0000", "$0,0.00000", "$0,0.000000", "$0,0.0000000", "$0,0.00000000", "$0,0.000000000", "$0,0.0000000000"];
PrivatePropFormats.moneyFormatGeneric = "$0,0.[0000000000000000000000000]";
PrivatePropFormats.percentFormat = ["0,0%", "0,0%", "0,0%", "0,0.0%", "0,0.00%", "0,0.000%", "0,0.0000%", "0,0.00000%", "0,0.000000%", "0,0.0000000%", "0,0.00000000%"];
PrivatePropFormats.percentFormatGeneric = "0,0.[0000000000000000000000000]%";
PrivatePropFormats.wholeFormat = "0,0";
class PropFormatter {
    /**
     * Get a string representation of this property suitable for 'reading'
     * @param prop
     * @param propDef
     * @returns {string}
     */
    static formatForRead(prop, propDef) {
        if (prop === null || prop === undefined) {
            return '';
        }
        else {
            return PropFormatter.formatValueForRead(prop.value, propDef);
        }
    }
    static formatValueForRead(value, propDef) {
        let lang = null;
        // The locale from the browser is not reliable.  The Extender server pulls the browser's locale from the
        // agent string at logon time.  Use that with a fallback approach to find the best fit locale.
        // var localeTest = window.navigator.userLanguage || window.navigator.language;
        if (dialog_1.CatavoltApi.singleton.browserLocaleJson) {
            let browserLocale = JSON.parse(dialog_1.CatavoltApi.singleton.browserLocaleJson); // country/language/varient
            if (browserLocale.country) {
                let key = browserLocale.language + "-" + browserLocale.country.toLowerCase();
                if (moment().lang(key)) {
                    lang = key;
                }
            }
            if (!lang) {
                let x = moment().lang(browserLocale.language);
                if (moment().lang(browserLocale.language)) {
                    lang = browserLocale.language;
                }
            }
            if (!lang) {
                lang = "en";
            }
        }
        // let test = (moment as any).locales();
        if (value === null || value === undefined) {
            return '';
        }
        else if ((propDef && propDef.isCodeRefType) || value instanceof CodeRef) {
            return value.description;
        }
        else if ((propDef && propDef.isObjRefType) || value instanceof ObjectRef) {
            return value.description;
        }
        else if ((propDef && propDef.isDateTimeType)) {
            if (!lang) {
                return value.toString();
            }
            else {
                return moment(value).locale(lang).format("lll");
                // return moment(value as Date).format("lll");
            }
        }
        else if ((propDef && propDef.isDateType) || value instanceof Date) {
            if (!lang) {
                return value.toLocaleDateString();
            }
            else {
                return moment(value).locale(lang).format("L");
                // return moment(value as Date).format("L");
            }
        }
        else if ((propDef && propDef.isTimeType) || value instanceof util_1.TimeValue) {
            if (!lang) {
                return moment(value).format("LT");
            }
            else {
                return moment(value).locale(lang).format("LT");
                // return moment(value as TimeValue).format("LT");
            }
        }
        else if ((propDef && propDef.isPasswordType)) {
            return value.replace(/./g, "*");
        }
        else if ((propDef && propDef.isListType) || Array.isArray(value)) {
            return value.reduce((prev, current) => {
                return ((prev ? prev + ', ' : '') + PropFormatter.formatValueForRead(current, null));
            }, '');
        }
        else {
            return PropFormatter.toString(value, propDef);
        }
    }
    /**
     * Get a string representation of this property suitable for 'writing'
     * @param prop
     * @param propDef
     * @returns {string}
     */
    static formatForWrite(prop, propDef) {
        if (prop === null || prop === undefined
            || prop.value === null || prop.value === undefined) {
            return null;
        }
        else if ((propDef && propDef.isCodeRefType) || prop.value instanceof CodeRef) {
            return prop.value.description;
        }
        else if ((propDef && propDef.isObjRefType) || prop.value instanceof ObjectRef) {
            return prop.value.description;
        }
        else {
            return PropFormatter.toStringWrite(prop.value, propDef);
        }
    }
    /**
     * Attempt to construct (or preserve) the appropriate data type given primitive (or already constructed) value.
     * @param value
     * @param propDef
     * @returns {any}
     */
    static parse(value, propDef) {
        var propValue = value;
        if (propDef.isDecimalType) {
            propValue = Number(value);
        }
        else if (propDef.isLongType) {
            propValue = Number(value);
        }
        else if (propDef.isBooleanType) {
            if (typeof value === 'string') {
                propValue = value !== 'false';
            }
            else {
                propValue = !!value;
            }
        }
        else if (propDef.isDateType) {
            //this could be a DateValue, a Date, or a string
            if (value instanceof util_1.DateValue) {
                propValue = value;
            }
            else if (typeof value === 'object') {
                propValue = new util_1.DateValue(value);
            }
            else {
                //parse as local time
                propValue = new util_1.DateValue(moment(value).toDate());
            }
        }
        else if (propDef.isDateTimeType) {
            //this could be a DateTimeValue, a Date, or a string
            if (value instanceof util_1.DateTimeValue) {
                propValue = value;
            }
            else if (typeof value === 'object') {
                propValue = new util_1.DateTimeValue(value);
            }
            else {
                //parse as local time
                propValue = new util_1.DateTimeValue(moment(value).toDate());
            }
        }
        else if (propDef.isTimeType) {
            propValue = value instanceof util_1.TimeValue ? value : util_1.TimeValue.fromString(value);
        }
        return propValue;
    }
    static resetFormats() {
        PropFormatter.decimalFormat = PrivatePropFormats.decimalFormat.slice(0);
        PropFormatter.decimalFormatGeneric = PrivatePropFormats.decimalFormatGeneric;
        PropFormatter.moneyFormat = PrivatePropFormats.moneyFormat.slice(0);
        PropFormatter.moneyFormatGeneric = PrivatePropFormats.moneyFormatGeneric;
        PropFormatter.percentFormat = PrivatePropFormats.percentFormat.slice(0);
        PropFormatter.percentFormatGeneric = PrivatePropFormats.decimalFormatGeneric;
        PropFormatter.wholeFormat = PrivatePropFormats.wholeFormat;
    }
    static toString(o, propDef) {
        return PropFormatter.toStringRead(o, propDef);
    }
    /**
     * Render this value as a string
     * @param o
     * @param propDef
     * @returns {any}
     */
    static toStringRead(o, propDef) {
        if (typeof o === 'number') {
            if (propDef && propDef.semanticType !== "DATA_UNFORMATTED_NUMBER") {
                if (propDef.isMoneyType) {
                    let f = propDef.displayScale < this.moneyFormat.length ? this.moneyFormat[propDef.displayScale] : this.moneyFormatGeneric;
                    // If there is a currency symbol, remove it noting it's position pre/post
                    // Necesary because numeral will replace $ with the symbol based on the locale of the browser.
                    // This may be desired down the road, but for now, the server provides the symbol to use.
                    let atStart = f.length > 0 && f[0] === '$';
                    let atEnd = f.length > 0 && f[f.length - 1] === '$';
                    if (dialog_1.CatavoltApi.singleton.currencySymbol) {
                        f = f.replace("$", ""); // Format this as a number, and slam in Extender currency symbol
                        var formatted = numeral(o).format(f);
                        if (atStart)
                            formatted = dialog_1.CatavoltApi.singleton.currencySymbol + formatted;
                        if (atEnd)
                            formatted = formatted + dialog_1.CatavoltApi.singleton.currencySymbol;
                    }
                    else {
                        formatted = numeral(o).format(f); // Should substitute browsers locale currency symbol
                    }
                    return formatted;
                }
                else if (propDef.isPercentType) {
                    let f = propDef.displayScale < this.percentFormat.length ? this.percentFormat[propDef.displayScale] : this.percentFormatGeneric;
                    return numeral(o).format(f); // numeral accomplishs * 100, relevant if we use some other symbol
                }
                else if (propDef.isIntType || propDef.isLongType) {
                    return numeral(o).format(this.wholeFormat);
                }
                else if (propDef.isDecimalType || propDef.isDoubleType) {
                    let f = propDef.displayScale < this.decimalFormat.length ? this.decimalFormat[propDef.displayScale] : this.decimalFormatGeneric;
                    return numeral(o).format(f);
                }
            }
            else {
                return String(o);
            }
        }
        else if (typeof o === 'object') {
            if (o instanceof Date) {
                return o.toISOString();
            }
            else if (o instanceof util_1.DateValue) {
                return o.dateObj.toISOString();
            }
            else if (o instanceof util_1.DateTimeValue) {
                return o.dateObj.toISOString();
            }
            else if (o instanceof util_1.TimeValue) {
                return o.toString();
            }
            else if (o instanceof CodeRef) {
                return o.toString();
            }
            else if (o instanceof ObjectRef) {
                return o.toString();
            }
            else if (o instanceof GpsReadingProperty) {
                return o.toString();
            }
            else if (o instanceof MapLocationProperty) {
                return o.toString();
            }
            else {
                return String(o);
            }
        }
        else {
            return String(o);
        }
    }
    static toStringWrite(o, propDef) {
        if (typeof o === 'number' && propDef) {
            let s = numeral(100);
            if (propDef.isMoneyType) {
                return o.toFixed(2);
            }
            else if (propDef.isIntType || propDef.isLongType) {
                return o.toFixed(0);
            }
            else if (propDef.isDecimalType || propDef.isDoubleType) {
                return o.toFixed(Math.max(2, (o.toString().split('.')[1] || []).length));
            }
        }
        else {
            return PropFormatter.toStringRead(o, propDef);
        }
    }
}
// For numeral format options, see: http://numeraljs.com/
// Default format for money at varying decimal lengths.
PropFormatter.decimalFormat = PrivatePropFormats.decimalFormat.slice(0);
PropFormatter.decimalFormatGeneric = PrivatePropFormats.decimalFormatGeneric;
PropFormatter.moneyFormat = PrivatePropFormats.moneyFormat.slice(0);
PropFormatter.moneyFormatGeneric = PrivatePropFormats.moneyFormatGeneric;
PropFormatter.percentFormat = PrivatePropFormats.percentFormat.slice(0);
PropFormatter.percentFormatGeneric = PrivatePropFormats.decimalFormatGeneric;
PropFormatter.wholeFormat = PrivatePropFormats.wholeFormat;
exports.PropFormatter = PropFormatter;
/**
 * In the same way that a {@link PropertyDef} describes a {@link Property}, a RecordDef describes an {@link Record}.
 * It is composed of {@link PropertyDef}s while the {@link Record} is composed of {@link Property}s.
 * In other words it describes the structure or makeup of a row or record, but does not contain the data values themselves.
 * The corresponding {@link Record} contains the actual values.
 */
class RecordDef {
    get propCount() {
        return this.propertyDefs.length;
    }
    propDefAtName(name) {
        var propDef = null;
        this.propertyDefs.some((p) => {
            if (p.propertyName === name) {
                propDef = p;
                return true;
            }
            return false;
        });
        return propDef;
    }
    get propNames() {
        return this.propertyDefs.map((p) => {
            return p.propertyName;
        });
    }
}
exports.RecordDef = RecordDef;
class RedirectionUtil {
    static isRedirection(o) {
        return [TypeNames.DialogRedirectionTypeName,
            TypeNames.NullRedirectionTypeName,
            TypeNames.WebRedirectionTypeName,
            TypeNames.WorkbenchRedirectionTypeName
        ].some(n => n === o['type']);
    }
    static isDialogRedirection(o) {
        return o['type'] === TypeNames.DialogRedirectionTypeName;
    }
    static isNullRedirection(o) {
        return o['type'] === TypeNames.NullRedirectionTypeName;
    }
    static isWebRedirection(o) {
        return o['type'] === TypeNames.WebRedirectionTypeName;
    }
    static isWorkbenchRedirection(o) {
        return o['type'] === TypeNames.WorkbenchRedirectionTypeName;
    }
}
exports.RedirectionUtil = RedirectionUtil;
class Stream extends View {
}
exports.Stream = Stream;
/**
 * A text template containing substitution parameters that is instantiated at presentation time and filled with business values.
 */
class SubstitutionCellValue extends CellValue {
    constructor(style, value) {
        super(style);
        this.value = value;
    }
}
exports.SubstitutionCellValue = SubstitutionCellValue;
/**
 * A purely declarative type. This object has no additional properties.
 */
class TabCellValue extends CellValue {
    constructor(style) {
        super(style);
    }
}
exports.TabCellValue = TabCellValue;
class UserMessage {
    constructor(message, messageType, explanation, propertyNames) {
        this.message = message;
        this.messageType = messageType;
        this.explanation = explanation;
        this.propertyNames = propertyNames;
    }
}
exports.UserMessage = UserMessage;
class ViewDescriptor {
    constructor(id, name, title) {
        this.id = id;
        this.name = name;
        this.title = title;
    }
}
exports.ViewDescriptor = ViewDescriptor;
/*
    ***************************************************************************
    Begin Dialog classes implementation
    ***************************************************************************
 */
/**
 * Top-level class, representing a Catavolt 'Dialog' definition.
 * All Dialogs have a composite {@link View} definition along with a single record
 * or a list of records.  See {@Record}
 */
class Dialog {
    constructor() {
        //private/protected
        this._binaryCache = {};
        this._lastRefreshTime = new Date(0);
        this.children = [];
    }
    /* public methods */
    get catavolt() {
        return this._catavolt;
    }
    /**
     * Load a Binary property from a record
     * @param propName
     * @param record
     * @returns {any}
     */
    binaryAt(propName, record) {
        const prop = record.propAtName(propName);
        if (prop) {
            if (prop.value instanceof InlineBinaryRef) {
                const binRef = prop.value;
                return Promise.resolve(new EncodedBinary(binRef.inlineData, binRef.settings['mime-type']));
            }
            else if (prop.value instanceof ObjectBinaryRef) {
                const binRef = prop.value;
                if (binRef.settings['webURL']) {
                    return Promise.resolve(new UrlBinary(binRef.settings['webURL']));
                }
                else {
                    return this.readBinary(propName, record);
                }
            }
            else if (typeof prop.value === 'string') {
                return Promise.resolve(new UrlBinary(prop.value));
            }
            else if (prop.value instanceof EncodedBinary) {
                return Promise.resolve(prop.value);
            }
            else {
                return Promise.reject('No binary found at ' + propName);
            }
        }
        else {
            return Promise.reject('No binary found at ' + propName);
        }
    }
    destroy() {
        //@TODO
        //destroy this dialog
    }
    /**
     * Return the error associated with this dialog, if any
     * @returns {any}
     */
    get error() {
        if (this.hasError) {
            return this.view.exception;
        }
        else {
            return null;
        }
    }
    /**
     * Find a menu def on this dialog with the given actionId
     * @param actionId
     * @returns {Menu}
     */
    findMenuAt(actionId) {
        return this.view.findMenuAt(actionId);
    }
    /**
     * Get a string representation of this property suitable for 'reading'
     * @param propValue
     * @param propName
     * @returns {string}
     */
    formatForRead(prop, propName) {
        return PropFormatter.formatForRead(prop, this.propDefAtName(propName));
    }
    /**
     * Get a string representation of this property suitable for 'writing'
     * @param propValue
     * @param propName
     * @returns {string}
     */
    formatForWrite(prop, propName) {
        return PropFormatter.formatForWrite(prop, this.propDefAtName(propName));
    }
    /**
     * Returns whether or not this dialog loaded properly
     * @returns {boolean}
     */
    get hasError() {
        return this.view instanceof ErrorMessage;
    }
    /**
     * Returns whether or not this Form is destroyed
     * @returns {boolean}
     */
    get isDestroyed() {
        return this.dialogMode === DialogModeEnum.DESTROYED || this.isAnyChildDestroyed;
    }
    /**
     * Returns whether or not the data in this dialog is out of date
     * @returns {boolean}
     */
    get isRefreshNeeded() {
        return this._lastRefreshTime.getTime() < this.catavolt.dataLastChangedTime.getTime();
    }
    /**
     * Get the last time this dialog's data was refreshed
     * @returns {Date}
     */
    get lastRefreshTime() {
        return this._lastRefreshTime;
    }
    /**
     * @param time
     */
    set lastRefreshTime(time) {
        this._lastRefreshTime = time;
    }
    /**
     * Get the all {@link Menu}'s associated with this dialog
     * @returns {Array<Menu>}
     */
    get menu() {
        return this.view.menu;
    }
    openView(targetViewDescriptor) {
        return this.catavolt.dialogApi.changeView(this.tenantId, this.sessionId, this.id, targetViewDescriptor.id)
            .then((dialog) => {
            //any new dialog needs to be initialized with the Catavolt object
            dialog.initialize(this.catavolt);
            this.updateSettingsWithNewDialogProperties(dialog.referringObject);
            return dialog;
        });
    }
    ;
    /**
     * Get the title of this dialog
     * @returns {string}
     */
    get paneTitle() {
        let title = this.view.findTitle();
        if (!title)
            title = this.description;
        return title;
    }
    /**
     * Parses a value to prepare for 'writing' back to the server
     * @param formattedValue
     * @param propName
     * @returns {any}
     */
    parseValue(formattedValue, propName) {
        return PropFormatter.parse(formattedValue, this.propDefAtName(propName));
    }
    /**
     * Get the propery definition for a property name
     * @param propName
     * @returns {PropDef}
     */
    propDefAtName(propName) {
        return this.recordDef.propDefAtName(propName);
    }
    /**
     * Read all the Binary values in this {@link Record}
     * @param record
     * @returns {Future<Array<Try<Binary>>>}
     */
    readBinaries(record) {
        return Promise.all(this.recordDef.propertyDefs.filter((propDef) => {
            return propDef.isBinaryType;
        }).map((propDef) => {
            return this.readBinary(propDef.propertyName, record);
        }));
    }
    get parentDialog() {
        return this._parentDialog;
    }
    /**
     * Get the all {@link ViewDescriptor}'s associated with this Form
     * @returns {Array<ViewDescriptor>}
     */
    get viewDescs() {
        return this.availableViews;
    }
    /* @TODO */
    writeAttachment(attachment) {
        /*
       return DialogService.addAttachment(this.dialogRedirection.dialogHandle, attachment, this.session);
       */
        return Promise.resolve(null);
    }
    writeAttachments(record) {
        return Promise.all(record.properties.filter((prop) => {
            return prop.value instanceof Attachment;
        }).map((prop) => {
            const attachment = prop.value;
            return this.writeAttachment(attachment);
        }));
    }
    /**
     * Write all Binary values in this {@link Record} back to the server
     * @param record
     * @returns {Future<Array<Try<XWritePropertyResult>>>}
     */
    /* @TODO */
    writeBinaries(record) {
        /*return Promise.all(
            record.properties.filter((prop: Property) => {
                return this.propDefAtName(prop.name).isBinaryType;
            }).map((prop: Property) => {
                let writePromise:Promise<XWritePropertyResult> = Promise.resolve({} as XWritePropertyResult);
                if (prop.value) {
                    let ptr: number = 0;
                    const encBin: EncodedBinary = prop.value as EncodedBinary;
                    const data = encBin.data;
                    while (ptr < data.length) {
                        const boundPtr = (ptr: number) => {
                            writePromise = writePromise.then((prevResult) => {
                                const encSegment: string = (ptr + Dialog.CHAR_CHUNK_SIZE) <= data.length ? data.substr(ptr, Dialog.CHAR_CHUNK_SIZE) : data.substring(ptr);
                                return DialogService.writeProperty(this.paneDef.dialogRedirection.dialogHandle, prop.name, encSegment, ptr != 0, this.session);
                            });
                        }
                        boundPtr(ptr);
                        ptr += Dialog.CHAR_CHUNK_SIZE;
                    }
                } else {
                    // This is a delete
                    writePromise = writePromise.then((prevResult) => {
                        return DialogService.writeProperty(this.paneDef.dialogRedirection.dialogHandle, prop.name, null, false, this.sessionContext);
                    });
                }
                return writePromise;
            })
        );*/
        return Promise.resolve(null);
    }
    initialize(catavolt) {
        this._catavolt = catavolt;
        if (this.children) {
            this.children.forEach((child) => {
                child._parentDialog = this;
                child.initialize(catavolt);
            });
        }
    }
    /**
     * Perform this action associated with the given Menu on this dialog.
     * The targets array is expected to be an array of object ids.
     * @param {Menu} menu
     * @param {ActionParameters} actionParams
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    invokeMenuAction(menu, actionParams) {
        return this.catavolt.dialogApi.performAction(this.catavolt.session.tenantId, this.catavolt.session.id, this.id, menu.actionId, actionParams).then((result) => {
            if (RedirectionUtil.isRedirection(result)) {
                //@TODO - update relevant referring dialog settings on 'this' dialog
                this.updateSettingsWithNewDialogProperties(result.referringObject);
                //@TODO -use 'isLocalRefreshNeeded' instead of this - needs to be added to the Dialog API
                if (result.referringObject && result.referringObject['dialogProperties']) {
                    const dialogProps = result.referringObject['dialogProperties'];
                    if ((dialogProps['localRefresh'] && dialogProps['localRefresh'] === 'true' ||
                        dialogProps['globalRefresh'] && dialogProps['globalRefresh'] === 'true')) {
                        this.catavolt.dataLastChangedTime = new Date();
                    }
                    //@TODO - also, this check should go away - we will rely on 'isLocalRefreshNeeded' exclusively
                }
                else if (RedirectionUtil.isNullRedirection(result)) {
                    this.catavolt.dataLastChangedTime = new Date();
                }
            }
            else {
                this.catavolt.dataLastChangedTime = new Date();
            }
            return result;
        });
    }
    //@TODO
    readBinary(propName, record) {
        /*
        let seq: number = 0;
        let encodedResult: string = '';
        let inProgress: string = '';
        let f: (XReadPropertyResult) => Promise<Binary> = (result: XReadPropertyResult) => {
            if (result.hasMore) {
                inProgress += atob(result.data);  // If data is in multiple loads, it must be decoded/built/encoded
                return DialogService.readEditorProperty(this.paneDef.dialogRedirection.dialogHandle,
                    propName, ++seq, Dialog.BINARY_CHUNK_SIZE, this.sessionContext).bind(f);
            } else {
                if (inProgress) {
                    inProgress += atob(result.data);
                    encodedResult = btoa(inProgress);
                } else {
                    encodedResult = result.data;
                }
                return Promise.resolve<Binary>(new EncodedBinary(encodedResult));
            }
        }
        return DialogService.readEditorProperty(this.paneDef.dialogRedirection.dialogHandle,
            propName, seq, Dialog.BINARY_CHUNK_SIZE, this.sessionContext).bind(f);
            */
        return Promise.resolve(null);
    }
    ;
    updateSettingsWithNewDialogProperties(referringObject) {
        if (referringObject) {
            if (referringObject.isDialogReferrer()) {
                //@TODO - remove the uppercase conversion once all DialogModes come back from server as uppercase
                this.dialogMode = referringObject.dialogMode.toUpperCase();
            }
        }
    }
    /**
     * @private
     * @returns {boolean}
     */
    get isAnyChildDestroyed() {
        return this.children && this.children.some((dialog) => {
            return dialog.isDestroyed;
        });
    }
}
//statics
Dialog.BINARY_CHUNK_SIZE = 256 * 1024; //size in  byes for 'read' operation
Dialog.CHAR_CHUNK_SIZE = 128 * 1000; //size in chars for encoded 'write' operation
exports.Dialog = Dialog;
/**
 * PanContext Subtype that represents an 'Editor Dialog'.
 * An 'Editor' represents and is backed by a single Record and Record definition.
 * See {@link Record} and {@link RecordDef}.
 */
class EditorDialog extends Dialog {
    /**
     * Get the current buffered record
     * @returns {RecordBuffer}
     */
    get buffer() {
        if (!this._buffer) {
            this._buffer = new RecordBuffer(NullRecord.singleton);
        }
        return this._buffer;
    }
    changeViewMode(viewMode) {
        return this.catavolt.dialogApi.changeMode(this.tenantId, this.sessionId, this.id, viewMode)
            .then((dialog) => {
            //any new dialog needs to be initialized with the Catavolt object
            dialog.initialize(this.catavolt);
            this.updateSettingsWithNewDialogProperties(dialog.referringObject);
            return dialog;
        });
    }
    /**
     * Get the associated entity record
     * @returns {Record}
     */
    get record() {
        return this._buffer.toRecord();
    }
    /**
     * Get the current version of the entity record, with any pending changes present
     * @returns {Record}
     */
    get recordNow() {
        return this.record;
    }
    /**
     * Get the possible values for a 'constrained value' property
     * @param propName
     * @returns {Future<Array<any>>}
     */
    //@TODO
    getAvailableValues(propName) {
        /*
         return DialogService.getAvailableValues(this.paneDef.dialogHandle, propName,
         this.buffer.afterEffects(), this.sessionContext).map((valuesResult:XGetAvailableValuesResult)=> {
         return valuesResult.list;
         });
         */
        return Promise.resolve(null);
    }
    /**
     * Returns whether or not this cell definition contains a binary value
     * @param cellValueDef
     * @returns {PropDef|boolean}
     */
    isBinary(cellValue) {
        var propDef = this.propDefAtName(cellValue.propertyName);
        return propDef && (propDef.isBinaryType || (propDef.isURLType && cellValue.isInlineMediaStyle));
    }
    /**
     * Returns whether or not the buffers contain valid data via a successful read operation.
     * @returns {boolean}
     */
    get isFirstReadComplete() {
        return this._isFirstReadComplete;
    }
    /**
     * Returns whether or not this Editor is in 'read' mode
     * @returns {boolean}
     */
    get isReadMode() {
        return this.viewMode === ViewModeEnum.READ;
    }
    /**
     * Returns whether or not this property is read-only
     * @param propName
     * @returns {boolean}
     */
    isReadModeFor(propName) {
        if (!this.isReadMode) {
            var propDef = this.propDefAtName(propName);
            return !propDef || !propDef.writeAllowed || !propDef.writeEnabled;
        }
        return true;
    }
    /**
     * Returns whether or not this cell definition contains a binary value that should be treated as a signature control
     * @param cellValueDef
     * @returns {PropDef|boolean}
     */
    isSignature(cellValueDef) {
        var propDef = this.propDefAtName(cellValueDef.propertyName);
        return this.isBinary(cellValueDef) && propDef.isSignatureType;
    }
    /**
     * Returns whether or not this property is 'writable'
     * @returns {boolean}
     */
    get isWriteMode() {
        return this.viewMode === ViewModeEnum.WRITE;
    }
    /**
     * Perform the action associated with the given Menu on this EditorDialog
     * Given that the Editor could possibly be destroyed as a result of this action,
     * any provided pending writes will be saved if present.
     * @param {Menu} menu
     * @param {Record} pendingWrites
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    performMenuAction(menu, pendingWrites) {
        return this.invokeMenuAction(menu, { pendingWrites: pendingWrites, type: TypeNames.ActionParametersTypeName }).then(result => {
            return result;
        });
    }
    /**
     * Properties whose {@link PropDef.canCauseSideEffects} value is true, may change other underlying values in the model.
     * This method will update those underlying values, given the property name that is changing, and the new value.
     * This is frequently used with {@link EditorContext.getAvailableValues}.  When a value is seleted, other properties
     * available values may change. (i.e. Country, State, City dropdowns)
     * @param propertyName
     * @param value
     * @returns {Future<null>}
     */
    //@TODO
    processSideEffects(propertyName, value) {
        /*
         var sideEffectsFr:Future<Record> = DialogService.processSideEffects(this.paneDef.dialogHandle,
         this.sessionContext, propertyName, value, this.buffer.afterEffects()).map((changeResult:XPropertyChangeResult)=> {
         return changeResult.sideEffects ? changeResult.sideEffects.record : new NullRecord();
         });

         return sideEffectsFr.map((sideEffectsRec:Record)=> {
         var originalProps = this.buffer.before.props;
         var userEffects = this.buffer.afterEffects().props;
         var sideEffects = sideEffectsRec.props;
         sideEffects = sideEffects.filter((prop:Prop)=> {
         return prop.name !== propertyName;
         });
         this._buffer = RecordBuffer.createRecordBuffer(this.buffer.objectId,
         RecordUtil.union(originalProps, sideEffects),
         RecordUtil.union(originalProps, RecordUtil.union(userEffects, sideEffects)));
         return null;
         });
         */
        return Promise.resolve(null);
    }
    /**
     * Read (load) the {@link Record} assocated with this Editor
     * The record must be read at least once to initialize the Context
     * @returns {Future<Record>}
     */
    read() {
        return this.catavolt.dialogApi.getRecord(this.tenantId, this.sessionId, this.id)
            .then((record) => {
            this._isFirstReadComplete = true;
            this.initBuffer(record);
            this.lastRefreshTime = new Date();
            return record;
        });
    }
    /**
     * Set the value of a property in this {@link Record}.
     * Values may be already constructed target types (CodeRef, TimeValue, Date, etc.)
     * or primitives, in which case the values will be parsed and objects constructed as necessary.
     * @param name
     * @param value
     * @returns {any}
     */
    setPropValue(name, value) {
        const propDef = this.propDefAtName(name);
        let parsedValue = null;
        if (propDef) {
            parsedValue = (value !== null && value !== undefined) ? this.parseValue(value, propDef.propertyName) : null;
            this.buffer.setValue(propDef.propertyName, parsedValue);
        }
        return parsedValue;
    }
    /**
     * Set a binary property from a string formatted as a 'data url'
     * See {@link https://en.wikipedia.org/wiki/Data_URI_scheme}
     * @param name
     * @param dataUrl
     */
    setBinaryPropWithDataUrl(name, dataUrl) {
        if (dataUrl) {
            const urlObj = new util_1.DataUrl(dataUrl);
            this.setBinaryPropWithEncodedData(name, urlObj.data, urlObj.mimeType);
        }
        else {
            this.setPropValue(name, null); // Property is being deleted/cleared
        }
    }
    /**
     * Set a binary property with base64 encoded data
     * @param name
     * @param encodedData
     * @param mimeType
     */
    setBinaryPropWithEncodedData(name, encodedData, mimeType) {
        const propDef = this.propDefAtName(name);
        if (propDef) {
            const value = new EncodedBinary(encodedData, mimeType);
            this.buffer.setValue(propDef.propertyName, value);
        }
    }
    /**
     * Write this record (i.e. {@link Record}} back to the server
     * @returns {Promise<Record | Redirection>}
     */
    write() {
        let deltaRec = this.buffer.afterEffects();
        /* Write the 'special' props first */
        return this.writeBinaries(deltaRec).then((binResult) => {
            return this.writeAttachments(deltaRec).then((atResult) => {
                /* Remove special property types before writing the actual record */
                deltaRec = this.removeSpecialProps(deltaRec);
                return this.catavolt.dialogApi.putRecord(this.tenantId, this.sessionId, this.id, deltaRec)
                    .then((result) => {
                    const now = new Date();
                    this.catavolt.dataLastChangedTime = now;
                    this.lastRefreshTime = now;
                    if (RedirectionUtil.isRedirection(result)) {
                        this.updateSettingsWithNewDialogProperties(result.referringObject);
                    }
                    else {
                        this.initBuffer(result);
                    }
                    return result;
                });
            });
        });
    }
    //Module level methods
    //Private methods
    /*
        @TODO
        Consider clone and deep copy here, to avoid potential ui side-effects
     */
    removeSpecialProps(record) {
        record.properties = record.properties.filter((prop) => {
            /* Remove the Binary(s) as they have been written seperately */
            return !this.propDefAtName(prop.name).isBinaryType;
        }).map((prop) => {
            /*
             Remove the Attachment(s) (as they have been written seperately) but replace
             the property value with the file name of the attachment prior to writing
             */
            if (prop.value instanceof Attachment) {
                const attachment = prop.value;
                return new Property(prop.name, attachment.name, prop.propertyType, prop.format, prop.annotations);
            }
            else {
                return prop;
            }
        });
        return record;
    }
    initBuffer(record) {
        this._buffer = record ? new RecordBuffer(record) : new RecordBuffer(NullRecord.singleton);
    }
}
exports.EditorDialog = EditorDialog;
/**
 * Dialog Subtype that represents a 'Query Dialog'.
 * A 'Query' represents and is backed by a list of Records and a single Record definition.
 * See {@link Record} and {@link RecordDef}.
 */
class QueryDialog extends Dialog {
    /**
     * Returns whether or not a column is of a binary type
     * @param columnDef
     * @returns {PropDef|boolean}
     */
    isBinary(column) {
        var propDef = this.propDefAtName(column.propertyName);
        return propDef && (propDef.isBinaryType || (propDef.isURLType && propDef.isInlineMediaStyle));
    }
    get defaultActionId() {
        return this._defaultActionId;
    }
    /**
     * Perform this action associated with the given Menu on this dialog.
     * The targets array is expected to be an array of object ids.
     * @param {Menu} menu
     * @param {Array<string>} targets
     * @returns {Promise<{actionId: string} | Redirection>}
     */
    performMenuAction(menu, targets) {
        return this.invokeMenuAction(menu, {
            targets: targets,
            type: TypeNames.ActionParametersTypeName
        }).then(result => {
            return result;
        });
    }
    /**
     * Perform a query
     * Note: {@link QueryScroller} is the preferred way to perform a query.
     * see {@link QueryContext.newScroller} and {@link QueryContext.setScroller}
     * @param maxRows
     * @param direction
     * @param fromObjectId
     * @returns {Future<RecordSet>}
     */
    query(maxRows, direction, fromObjectId) {
        const queryParams = fromObjectId ?
            {
                fetchDirection: direction,
                fetchMaxRecords: maxRows,
                fromBusinessId: fromObjectId,
                type: TypeNames.QueryParametersTypeName
            } :
            { fetchDirection: direction, fetchMaxRecords: maxRows, type: TypeNames.QueryParametersTypeName };
        return this.catavolt.dialogApi.getRecords(this.catavolt.session.tenantId, this.catavolt.session.id, this.id, queryParams)
            .then((recordSet) => {
            this.lastRefreshTime = new Date();
            this._defaultActionId = recordSet.defaultActionId;
            return recordSet;
        });
    }
    /**
     * Clear the QueryScroller's buffer and perform this query
     * @returns {Future<Array<Record>>}
     */
    refresh() {
        return this._scroller.refresh();
    }
    /**
     * Get the associated QueryScroller
     * @returns {QueryScroller}
     */
    get scroller() {
        if (!this._scroller) {
            this._scroller = this.newScroller();
        }
        return this._scroller;
    }
    /**
     * Creates a new QueryScroller with the given values
     * @param pageSize
     * @param firstObjectId
     * @param markerOptions
     * @returns {QueryScroller}
     */
    setScroller(pageSize, firstObjectId = null, markerOptions = [QueryMarkerOption.None]) {
        this._scroller = new QueryScroller(this, pageSize, firstObjectId, markerOptions);
        return this._scroller;
    }
    /**
     * Creates a new QueryScroller with default buffer size of 50
     * @returns {QueryScroller}
     */
    newScroller() {
        return this.setScroller(50, null, [QueryMarkerOption.None]);
    }
}
exports.QueryDialog = QueryDialog;
/**
 * *********************************
 */
class HasMoreQueryMarker extends NullRecord {
}
HasMoreQueryMarker.singleton = new HasMoreQueryMarker();
exports.HasMoreQueryMarker = HasMoreQueryMarker;
class IsEmptyQueryMarker extends NullRecord {
}
IsEmptyQueryMarker.singleton = new IsEmptyQueryMarker();
exports.IsEmptyQueryMarker = IsEmptyQueryMarker;
var QueryMarkerOption;
(function (QueryMarkerOption) {
    QueryMarkerOption[QueryMarkerOption["None"] = 0] = "None";
    QueryMarkerOption[QueryMarkerOption["IsEmpty"] = 1] = "IsEmpty";
    QueryMarkerOption[QueryMarkerOption["HasMore"] = 2] = "HasMore";
})(QueryMarkerOption = exports.QueryMarkerOption || (exports.QueryMarkerOption = {}));
class QueryScroller {
    constructor(_dialog, _pageSize, _firstObjectId, _markerOptions = []) {
        this._dialog = _dialog;
        this._pageSize = _pageSize;
        this._firstObjectId = _firstObjectId;
        this._markerOptions = _markerOptions;
        this.clear();
    }
    get buffer() {
        return this._buffer;
    }
    get bufferWithMarkers() {
        var result = util_1.ArrayUtil.copy(this._buffer);
        if (this.isComplete) {
            if (this._markerOptions.indexOf(QueryMarkerOption.IsEmpty) > -1) {
                if (this.isEmpty) {
                    result.push(IsEmptyQueryMarker.singleton);
                }
            }
        }
        else if (this._markerOptions.indexOf(QueryMarkerOption.HasMore) > -1) {
            if (result.length === 0) {
                result.push(HasMoreQueryMarker.singleton);
            }
            else {
                if (this._hasMoreBackward) {
                    result.unshift(HasMoreQueryMarker.singleton);
                }
                if (this._hasMoreForward) {
                    result.push(HasMoreQueryMarker.singleton);
                }
            }
        }
        return result;
    }
    get dialog() {
        return this._dialog;
    }
    get firstObjectId() {
        return this._firstObjectId;
    }
    get hasMoreBackward() {
        return this._hasMoreBackward;
    }
    get hasMoreForward() {
        return this._hasMoreForward;
    }
    get isComplete() {
        return !this._hasMoreBackward && !this._hasMoreForward;
    }
    get isCompleteAndEmpty() {
        return this.isComplete && this._buffer.length === 0;
    }
    get isEmpty() {
        return this._buffer.length === 0;
    }
    pageBackward() {
        if (!this._hasMoreBackward) {
            return Promise.resolve([]);
        }
        if (this._prevPagePromise) {
            this._prevPagePromise = this._prevPagePromise.then((recordSet) => {
                const fromObjectId = this._buffer.length === 0 ? null : this._buffer[0].id;
                return this._dialog.query(this._pageSize, QueryDirectionEnum.BACKWARD, fromObjectId);
            });
        }
        else {
            const fromObjectId = this._buffer.length === 0 ? null : this._buffer[0].id;
            this._prevPagePromise = this._dialog.query(this._pageSize, QueryDirectionEnum.BACKWARD, fromObjectId);
        }
        const beforeSize = this._buffer.length;
        return this._prevPagePromise.then((queryResult) => {
            let afterSize = beforeSize;
            this._hasMoreBackward = queryResult.hasMore;
            if (queryResult.records.length > 0) {
                const newBuffer = [];
                for (let i = queryResult.records.length - 1; i > -1; i--) {
                    newBuffer.push(queryResult.records[i]);
                }
                this._buffer.forEach((record) => {
                    newBuffer.push(record);
                });
                this._buffer = newBuffer;
                afterSize = this._buffer.length;
            }
            return queryResult.records;
        });
    }
    pageForward() {
        if (!this._hasMoreForward) {
            return Promise.resolve([]);
        }
        if (this._nextPagePromise) {
            this._nextPagePromise = this._nextPagePromise.then((recordSet) => {
                const fromObjectId = this._buffer.length === 0 ? null : this._buffer[this._buffer.length - 1].id;
                return this._dialog.query(this._pageSize, QueryDirectionEnum.FORWARD, fromObjectId);
            });
        }
        else {
            const fromObjectId = this._buffer.length === 0 ? null : this._buffer[this._buffer.length - 1].id;
            this._nextPagePromise = this._dialog.query(this._pageSize, QueryDirectionEnum.FORWARD, fromObjectId);
        }
        const beforeSize = this._buffer.length;
        return this._nextPagePromise.then((queryResult) => {
            let afterSize = beforeSize;
            this._hasMoreForward = queryResult.hasMore;
            if (queryResult.records.length > 0) {
                const newBuffer = [];
                this._buffer.forEach((record) => {
                    newBuffer.push(record);
                });
                queryResult.records.forEach((record) => {
                    newBuffer.push(record);
                });
                this._buffer = newBuffer;
                afterSize = this._buffer.length;
            }
            return queryResult.records;
        });
    }
    get pageSize() {
        return this._pageSize;
    }
    refresh() {
        this.clear();
        return this.pageForward().then((recordList) => {
            if (recordList.length > 0) {
                this._firstResultOid = recordList[0].id;
            }
            return recordList;
        });
    }
    trimFirst(n) {
        const newBuffer = [];
        for (let i = n; i < this._buffer.length; i++) {
            newBuffer.push(this._buffer[i]);
        }
        this._buffer = newBuffer;
        this._hasMoreBackward = true;
    }
    trimLast(n) {
        var newBuffer = [];
        for (let i = 0; i < this._buffer.length - n; i++) {
            newBuffer.push(this._buffer[i]);
        }
        this._buffer = newBuffer;
        this._hasMoreForward = true;
    }
    clear() {
        this._hasMoreBackward = !!this._firstObjectId;
        this._hasMoreForward = true;
        this._buffer = [];
        this._firstResultOid = null;
    }
}
exports.QueryScroller = QueryScroller;
class ErrorDialog extends Dialog {
}
exports.ErrorDialog = ErrorDialog;
/* DialogMode */
var DialogModeEnum;
(function (DialogModeEnum) {
    DialogModeEnum["COPY"] = "COPY";
    DialogModeEnum["CREATE"] = "CREATE";
    DialogModeEnum["READ"] = "READ";
    DialogModeEnum["UPDATE"] = "UPDATE";
    DialogModeEnum["DESTROYED"] = "DESTROYED";
    DialogModeEnum["DELETE"] = "DELETE";
    DialogModeEnum["LIST"] = "LIST";
})(DialogModeEnum = exports.DialogModeEnum || (exports.DialogModeEnum = {}));
var QueryDirectionEnum;
(function (QueryDirectionEnum) {
    QueryDirectionEnum["FORWARD"] = "FORWARD";
    QueryDirectionEnum["BACKWARD"] = "BACKWARD";
})(QueryDirectionEnum = exports.QueryDirectionEnum || (exports.QueryDirectionEnum = {}));
var ViewModeEnum;
(function (ViewModeEnum) {
    ViewModeEnum["READ"] = "READ";
    ViewModeEnum["WRITE"] = "WRITE";
})(ViewModeEnum = exports.ViewModeEnum || (exports.ViewModeEnum = {}));
/* Type descriminators */
var TypeNames;
(function (TypeNames) {
    TypeNames["ActionParametersTypeName"] = "hxgn.api.dialog.ActionParameters";
    TypeNames["AppWindowTypeName"] = "hxgn.api.dialog.AppWindow";
    TypeNames["BarcodeScan"] = "hxgn.api.dialog.BarcodeScan";
    TypeNames["CalendarTypeName"] = "hxgn.api.dialog.Calendar";
    TypeNames["CodeRefTypeName"] = "hxgn.api.dialog.CodeRef";
    TypeNames["DetailsTypeName"] = "hxgn.api.dialog.Details";
    TypeNames["DialogTypeName"] = "hxgn.api.dialog.Dialog";
    TypeNames["DialogMessageTypeName"] = "hxgn.api.dialog.DialogMessage";
    TypeNames["DialogRedirectionTypeName"] = "hxgn.api.dialog.DialogRedirection";
    TypeNames["EditorDialogTypeName"] = "hxgn.api.dialog.EditorDialog";
    TypeNames["FormTypeName"] = "hxgn.api.dialog.Form";
    TypeNames["GpsReadingTypeName"] = "hxgn.api.dialog.GpsReading";
    TypeNames["GpsReadingPropertyTypeName"] = "hxgn.api.dialog.GpsReadingProperty";
    TypeNames["GraphTypeName"] = "hxgn.api.dialog.Graph";
    TypeNames["ListTypeName"] = "hxgn.api.dialog.List";
    TypeNames["LoginTypeName"] = "hxgn.api.dialog.Login";
    TypeNames["MapTypeName"] = "hxgn.api.dialog.Map";
    TypeNames["MapLocationTypeName"] = "hxgn.api.dialog.MapLocation";
    TypeNames["MapLocationPropertyTypeName"] = "hxgn.api.dialog.MapLocationProperty";
    TypeNames["NullRedirectionTypeName"] = "hxgn.api.dialog.NullRedirection";
    TypeNames["ObjectRefTypeName"] = "hxgn.api.dialog.ObjectRef";
    TypeNames["PropertyTypeName"] = "hxgn.api.dialog.Property";
    TypeNames["QueryDialogTypeName"] = "hxgn.api.dialog.QueryDialog";
    TypeNames["QueryParametersTypeName"] = "hxgn.api.dialog.QueryParameters";
    TypeNames["RecordTypeName"] = "hxgn.api.dialog.Record";
    TypeNames["ReferringDialogTypeName"] = "hxgn.api.dialog.ReferringDialog";
    TypeNames["ReferringWorkbenchTypeName"] = "hxgn.api.dialog.ReferringWorkbench";
    TypeNames["SessionTypeName"] = "hxgn.api.dialog.Session";
    TypeNames["StreamTypeName"] = "hxgn.api.dialog.Stream";
    TypeNames["WebRedirectionTypeName"] = "hxgn.api.dialog.WebRedirection";
    TypeNames["WorkbenchTypeName"] = "hxgn.api.dialog.Workbench";
    TypeNames["WorkbenchRedirectionTypeName"] = "hxgn.api.dialog.WorkbenchRedirection";
})(TypeNames = exports.TypeNames || (exports.TypeNames = {}));
class ModelUtil {
    static classType(name) {
        return ModelUtil.classTypes[name];
    }
    static typeInstance(name) {
        const type = ModelUtil.classType(name);
        return type && new type;
    }
    static jsonToModel(obj, n = 0) {
        const indent = n * 4;
        if (Array.isArray(obj)) {
            //Log.debug(`${' '.repeat(indent)}=> Deserializing Array....`);
            return ModelUtil.deserializeArray(obj);
        }
        else {
            const objType = obj['type'];
            //Log.debug(`${' '.repeat(indent)}=> Deserializing ${objType}`);
            return new Promise((resolve, reject) => {
                //if the class has a fromJSON method, use it
                const classType = ModelUtil.classType(objType);
                if (classType && typeof classType.fromJSON === 'function') {
                    resolve(classType.fromJSON(obj));
                }
                else {
                    let newObj = ModelUtil.typeInstance(objType);
                    if (!newObj) {
                        const message = `ModelUtil::jsonToModel: no type constructor found for ${objType}: assuming interface`;
                        //Log.debug(message);
                        newObj = {}; //assume it's an interface
                    }
                    //otherwise, copy field values
                    Promise.all(Object.keys(obj).map(prop => {
                        const value = obj[prop];
                        //Log.debug(`${' '.repeat(indent)}prop: ${prop} is type ${typeof value}`);
                        if (value && typeof value === 'object') {
                            if (Array.isArray(value) || 'type' in value) {
                                return ModelUtil.jsonToModel(value, ++n).then(model => {
                                    ModelUtil.assignProp(prop, model, newObj, objType, indent);
                                });
                            }
                            else {
                                ModelUtil.assignProp(prop, value, newObj, objType, indent);
                                return Promise.resolve();
                            }
                        }
                        else {
                            ModelUtil.assignProp(prop, value, newObj, objType, indent);
                            return Promise.resolve();
                        }
                    })).then(result => {
                        resolve(newObj);
                    }).catch(error => reject(error));
                }
            });
        }
    }
    static modelToJson(obj, filterFn) {
        return util_1.ObjUtil.copyNonNullFieldsOnly(obj, {}, (prop) => {
            return prop.charAt(0) !== '_' && (!filterFn || filterFn(prop));
        });
    }
    static deserializeArray(array) {
        return Promise.all(array.map(value => {
            if (value && typeof value === 'object') {
                return ModelUtil.jsonToModel(value);
            }
            else {
                return Promise.resolve(value);
            }
        }));
    }
    static assignProp(prop, value, target, type, n) {
        try {
            if ('_' + prop in target) {
                target['_' + prop] = value;
                //Log.debug(`${' '.repeat(n)}Assigning private prop _${prop} = ${value}`);
            }
            else {
                //it may be public prop
                if (prop in target) {
                    //Log.debug(`${' '.repeat(n)}Assigning public prop ${prop} = ${value}`);
                }
                else {
                    //it's either a readonly prop or defined in an interface
                    //in which case it's will not already exist on the target object
                    //Log.debug(`${' '.repeat(n)}Defining ${prop} on target for ${type}`);
                }
                target[prop] = value;
            }
        }
        catch (error) {
            util_1.Log.error(`ModelUtil::assignProp: Failed to set prop: ${prop} on target: ${error}`);
        }
    }
}
ModelUtil.classTypes = {
    'hxgn.api.dialog.Annotation': DataAnnotation,
    'hxgn.api.dialog.AttributeCellValue': AttributeCellValue,
    'hxgn.api.dialog.TabCellValue': TabCellValue,
    'hxgn.api.dialog.BarcodeScan': BarcodeScan,
    'hxgn.api.dialog.Calendar': Calendar,
    'hxgn.api.dialog.CodeRef': CodeRef,
    'hxgn.api.dialog.Details': Details,
    'hxgn.api.dialog.DialogException': DialogException,
    'hxgn.api.dialog.EditorDialog': EditorDialog,
    'hxgn.api.dialog.ForcedLineCellValue': ForcedLineCellValue,
    'hxgn.api.dialog.Form': Form,
    'hxgn.api.dialog.GpsReading': GpsReading,
    'hxgn.api.dialog.GpsReadingProperty': GpsReadingProperty,
    'hxgn.api.dialog.MapLocation': MapLocation,
    'hxgn.api.dialog.MapLocationProperty': MapLocationProperty,
    'hxgn.api.dialog.Graph': Graph,
    'hxgn.api.dialog.InlineBinaryRef': InlineBinaryRef,
    'hxgn.api.dialog.LabelCellValue': LabelCellValue,
    'hxgn.api.dialog.List': List,
    'hxgn.api.dialog.Map': Map,
    'hxgn.api.dialog.Menu': Menu,
    'hxgn.api.dialog.ObjectBinaryRef': ObjectBinaryRef,
    'hxgn.api.dialog.ObjectRef': ObjectRef,
    'hxgn.api.dialog.Property': Property,
    'hxgn.api.dialog.PropertyDef': PropertyDef,
    'hxgn.api.dialog.QueryDialog': QueryDialog,
    'hxgn.api.dialog.Record': RecordImpl,
    'hxgn.api.dialog.RecordDef': RecordDef,
    'hxgn.api.dialog.ReferringDialog': ReferringDialog,
    'hxgn.api.dialog.ReferringWorkbench': ReferringWorkbench,
    'hxgn.api.dialog.Stream': Stream,
    'hxgn.api.dialog.SubstitutionCellValue': SubstitutionCellValue,
    'hxgn.api.dialog.ViewDescriptor': ViewDescriptor
};
exports.ModelUtil = ModelUtil;
/**
 * *********************************
 */
