"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("./util");
class PersistenceTools {
    static deleteAllDialogState(tenantId, userId, dialogId) {
        const dialog = this.readDialogState(tenantId, userId, dialogId);
        if (dialog) {
            this.deleteAllDialogStateFor(tenantId, userId, dialog);
        }
        else {
            this.deleteRedirectionState(tenantId, userId, dialogId);
        }
    }
    static deleteAllDialogStateFor(tenantId, userId, dialog) {
        const dialogChildren = dialog.children;
        if (dialogChildren) {
            for (let child of dialogChildren) {
                this.deleteAllDialogStateFor(tenantId, userId, child);
            }
        }
        this.deleteRedirectionState(tenantId, userId, dialog.id);
        this.deleteDialogState(tenantId, userId, dialog.id);
        this.deleteRecordSetState(tenantId, userId, dialog.id);
        this.deleteRecordState(tenantId, userId, dialog.id);
        this.deleteDialogParentState(tenantId, userId, dialog.id);
        this.deleteDialogAliasState(tenantId, userId, dialog.id);
        this.deleteDialogReferringAliasState(tenantId, userId, dialog.id);
    }
    static deleteAllState(tenantId, userId) {
        const keyCount = localStorage.length;
        for (var i = keyCount - 1; i > -1; --i) {
            const key = localStorage.key(i);
            util_1.Log.debug("Removing from localStorage: " + key);
            localStorage.removeItem(key);
        }
    }
    static deleteAllWorkbenchNavigation(tenantId, userId, navigationKey) {
        const previousNavigation = PersistenceTools.readNavigationState(tenantId, userId, navigationKey);
        if (previousNavigation) {
            PersistenceTools.deleteAllDialogState(tenantId, userId, previousNavigation.redirectionId);
        }
    }
    static deleteDialogAliasState(tenantId, userId, dialogId) {
        this.deletePersistentState(tenantId, userId, 'dialog.' + dialogId + '.alias');
    }
    static deleteDialogParentState(tenantId, userId, childId) {
        this.deletePersistentState(tenantId, userId, 'dialog.' + childId + '.parent');
    }
    static deleteDialogReferringAliasState(tenantId, userId, dialogId) {
        this.deletePersistentState(tenantId, userId, 'dialog.' + dialogId + '.referringAlias');
    }
    static deleteDialogState(tenantId, userId, dialogId) {
        this.deletePersistentState(tenantId, userId, 'dialog.' + dialogId);
    }
    static deleteNavigationState(tenantId, userId, navigationId) {
        this.deletePersistentState(tenantId, userId, 'navigation.' + navigationId);
    }
    static deletePersistentState(tenantId, userId, stateId) {
        const key = tenantId + '.' + userId + '.' + stateId;
        window.localStorage.removeItem(key);
    }
    static deleteRecordSetState(tenantId, userId, dialogId) {
        this.deletePersistentState(tenantId, userId, 'dialog.' + dialogId + '.recordSet');
    }
    static deleteRecordState(tenantId, userId, dialogId) {
        this.deletePersistentState(tenantId, userId, 'dialog.' + dialogId + '.record');
    }
    static deleteRedirectionState(tenantId, userId, redirectionId) {
        this.deletePersistentState(tenantId, userId, 'redirection.' + redirectionId);
    }
    static deleteSessionState(tenantId, userId) {
        this.deletePersistentState(tenantId, userId, 'session');
    }
    static findDialogState(tenantId, userId, dialogId) {
        const dialog = this.findRootDialogState(tenantId, userId, dialogId);
        if (!dialog) {
            return null;
        }
        return this.findDialogStateWithin(tenantId, userId, dialog, dialogId);
    }
    static findDialogStateWithin(tenantId, userId, dialog, targetId) {
        if (dialog && dialog.id === targetId) {
            return dialog;
        }
        const dialogChildren = dialog.children;
        if (dialogChildren) {
            for (let child of dialogChildren) {
                const nestedDialog = this.findDialogStateWithin(tenantId, userId, child, targetId);
                if (nestedDialog) {
                    return nestedDialog;
                }
            }
        }
        return null;
    }
    static findRecordProperty(record, propertyName) {
        for (let p of record.properties) {
            if (p.name === propertyName) {
                return p;
            }
        }
        return null;
    }
    static findRootDialogState(tenantId, userId, dialogId) {
        const dialog = this.readDialogState(tenantId, userId, dialogId);
        if (dialog) {
            return dialog;
        }
        const parentId = this.readDialogParentState(tenantId, userId, dialogId);
        if (parentId) {
            return this.findRootDialogState(tenantId, userId, parentId);
        }
        return null;
    }
    static isDeleteSession(path) {
        return path.length == 4 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS;
    }
    static isGetAvailableViews(path) {
        return path.length == 7 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.AVAILABLE_VIEWS;
    }
    static isGetDialog(path) {
        return path.length == 6 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS;
    }
    static isGetRecord(path) {
        return path.length == 7 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.RECORD;
    }
    static isGetRedirection(path) {
        return path.length == 6 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.REDIRECTIONS;
    }
    static isGetSelectedView(path) {
        return path.length == 7 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.SELECTED_VIEW;
    }
    static isGetSession(path) {
        return path.length == 4 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS;
    }
    static isGetTenant(path) {
        return path.length == 2 &&
            path[0] == PersistenceTools.TENANTS;
    }
    static isGetTenants(path) {
        return path.length == 1 &&
            path[0] == PersistenceTools.TENANTS;
    }
    static isGetViewMode(path) {
        return path.length == 7 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.VIEW_MODE;
    }
    static isGetWorkbenches(path) {
        return path.length == 5 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.WORKBENCHES;
    }
    static isPostAvailableValues(path) {
        return path.length == 9 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.RECORD &&
            path[8] == PersistenceTools.AVAILABLE_VALUES;
    }
    static isPostMenuAction(path) {
        return path.length == 8 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.ACTIONS;
    }
    static isPostRecords(path) {
        return path.length == 7 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.RECORDS;
    }
    static isPostSession(path) {
        return path.length == 3 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS;
    }
    static isPostWorkbenchAction(path) {
        return path.length == 8 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.WORKBENCHES &&
            path[6] == PersistenceTools.ACTIONS;
    }
    static isPutRecord(path) {
        return path.length == 7 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.RECORD;
    }
    static isPutSelectedView(path) {
        return path.length == 8 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.SELECTED_VIEW;
    }
    static isPutViewMode(path) {
        return path.length == 8 &&
            path[0] == PersistenceTools.TENANTS &&
            path[2] == PersistenceTools.SESSIONS &&
            path[4] == PersistenceTools.DIALOGS &&
            path[6] == PersistenceTools.VIEW_MODE;
    }
    static readDialogAliasState(tenantId, userId, dialogId) {
        return this.readPersistentState(tenantId, userId, 'dialog.' + dialogId + '.alias');
    }
    static readDialogParentState(tenantId, userId, childId) {
        return this.readPersistentState(tenantId, userId, 'dialog.' + childId + '.parent');
    }
    static readDialogReferringAliasState(tenantId, userId, dialogId) {
        return this.readPersistentState(tenantId, userId, 'dialog.' + dialogId + '.referringAlias');
    }
    static readDialogState(tenantId, userId, dialogId) {
        return this.readPersistentState(tenantId, userId, 'dialog.' + dialogId);
    }
    static readNavigationState(tenantId, userId, navigationId) {
        return this.readPersistentState(tenantId, userId, 'navigation.' + navigationId);
    }
    static readPersistentState(tenantId, userId, stateId) {
        const key = tenantId + '.' + userId + '.' + stateId;
        var json = window.localStorage.getItem(key);
        return json ? JSON.parse(json) : null;
    }
    static readRecordSetState(tenantId, userId, dialogId) {
        return this.readPersistentState(tenantId, userId, 'dialog.' + dialogId + '.recordSet');
    }
    static readRecordState(tenantId, userId, dialogId) {
        return this.readPersistentState(tenantId, userId, 'dialog.' + dialogId + '.record');
    }
    static readRedirectionState(tenantId, userId, redirectionId) {
        return this.readPersistentState(tenantId, userId, 'redirection.' + redirectionId);
    }
    static readSessionState(tenantId, userId) {
        return this.readPersistentState(tenantId, userId, 'session');
    }
    static updateRecordPropertyValue(record, propertyName, value) {
        for (let p of record.properties) {
            if (p.name === propertyName) {
                p.value = value;
                return true;
            }
        }
        return false;
    }
    static writeAllDialogParentState(tenantId, userId, parent) {
        const dialogChildren = parent.children;
        if (dialogChildren) {
            for (var i = 0; i < dialogChildren.length; i++) {
                const child = dialogChildren[i];
                this.writeDialogParentState(tenantId, userId, child, i, parent);
            }
        }
    }
    static writeDialogAliasState(tenantId, userId, dialogId, alias) {
        this.writePersistentState(tenantId, userId, 'dialog.' + dialogId + '.alias', alias);
    }
    static writeDialogParentState(tenantId, userId, child, childIndex, parent) {
        this.writePersistentState(tenantId, userId, 'dialog.' + child.id + '.parent', parent.id);
        const parentAlias = this.readDialogAliasState(tenantId, userId, parent.id);
        this.writeDialogAliasState(tenantId, userId, child.id, parentAlias + '_' + childIndex);
        this.writeAllDialogParentState(tenantId, userId, child);
    }
    static writeDialogReferringAliasState(tenantId, userId, dialogId, referringAlias) {
        this.writePersistentState(tenantId, userId, 'dialog.' + dialogId + '.referringAlias', referringAlias);
    }
    static writeDialogState(tenantId, userId, dialog) {
        this.writePersistentState(tenantId, userId, 'dialog.' + dialog.id, dialog);
    }
    static writePersistentState(tenantId, userId, stateId, state) {
        const key = tenantId + '.' + userId + '.' + stateId;
        window.localStorage.setItem(key, JSON.stringify(state));
    }
    static writeRecordSetState(tenantId, userId, dialogId, recordSet) {
        this.writePersistentState(tenantId, userId, 'dialog.' + dialogId + '.recordSet', recordSet);
    }
    static writeRecordState(tenantId, userId, dialogId, record) {
        this.writePersistentState(tenantId, userId, 'dialog.' + dialogId + '.record', record);
    }
    static writeNavigationState(tenantId, userId, navigation) {
        this.writePersistentState(tenantId, userId, 'navigation.' + navigation.id, navigation);
    }
    static writeRedirectionState(tenantId, userId, redirection) {
        this.writePersistentState(tenantId, userId, 'redirection.' + redirection.id, redirection);
    }
    static writeSessionState(session) {
        this.writePersistentState(session.tenantId, session.userId, 'session', session);
    }
}
PersistenceTools.ACTIONS = 'actions';
PersistenceTools.AVAILABLE_VALUES = 'availableValues';
PersistenceTools.AVAILABLE_VIEWS = 'availableViews';
PersistenceTools.DIALOGS = 'dialogs';
PersistenceTools.RECORD = 'record';
PersistenceTools.RECORDS = 'records';
PersistenceTools.REDIRECTIONS = 'redirections';
PersistenceTools.SELECTED_VIEW = 'selectedView';
PersistenceTools.SESSIONS = 'sessions';
PersistenceTools.TENANTS = 'tenants';
PersistenceTools.VIEW_MODE = 'viewMode';
PersistenceTools.WORKBENCHES = 'workbenches';
exports.PersistenceTools = PersistenceTools;
