"use strict";
/**
 * Created by rburson on 8/28/17.
 */
Object.defineProperty(exports, "__esModule", { value: true });
class ArrayUtil {
    static copy(source) {
        return source.map((e) => {
            return e;
        });
    }
    static find(source, f) {
        var value = null;
        source.some((v) => {
            if (f(v)) {
                value = v;
                return true;
            }
            return false;
        });
        return value;
    }
}
exports.ArrayUtil = ArrayUtil;
/**
 * *****************************************************
 */
/*
 This implementation supports our ECMA 5.1 browser set, including IE9
 If we no longer need to support IE9, a TypedArray implementaion would be more efficient...
 */
class Base64 {
    static encode(input) {
        var output = "";
        var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
        var i = 0;
        input = Base64._utf8_encode(input);
        while (i < input.length) {
            chr1 = input.charCodeAt(i++);
            chr2 = input.charCodeAt(i++);
            chr3 = input.charCodeAt(i++);
            enc1 = chr1 >> 2;
            enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
            enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
            enc4 = chr3 & 63;
            if (isNaN(chr2)) {
                enc3 = enc4 = 64;
            }
            else if (isNaN(chr3)) {
                enc4 = 64;
            }
            output = output +
                Base64._keyStr.charAt(enc1) + Base64._keyStr.charAt(enc2) +
                Base64._keyStr.charAt(enc3) + Base64._keyStr.charAt(enc4);
        }
        return output;
    }
    static decode(input) {
        var output = "";
        var chr1, chr2, chr3;
        var enc1, enc2, enc3, enc4;
        var i = 0;
        input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");
        while (i < input.length) {
            enc1 = Base64._keyStr.indexOf(input.charAt(i++));
            enc2 = Base64._keyStr.indexOf(input.charAt(i++));
            enc3 = Base64._keyStr.indexOf(input.charAt(i++));
            enc4 = Base64._keyStr.indexOf(input.charAt(i++));
            chr1 = (enc1 << 2) | (enc2 >> 4);
            chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
            chr3 = ((enc3 & 3) << 6) | enc4;
            output = output + String.fromCharCode(chr1);
            if (enc3 != 64) {
                output = output + String.fromCharCode(chr2);
            }
            if (enc4 != 64) {
                output = output + String.fromCharCode(chr3);
            }
        }
        output = Base64._utf8_decode(output);
        return output;
    }
    static _utf8_encode(s) {
        s = s.replace(/\r\n/g, "\n");
        var utftext = "";
        for (var n = 0; n < s.length; n++) {
            var c = s.charCodeAt(n);
            if (c < 128) {
                utftext += String.fromCharCode(c);
            }
            else if ((c > 127) && (c < 2048)) {
                utftext += String.fromCharCode((c >> 6) | 192);
                utftext += String.fromCharCode((c & 63) | 128);
            }
            else {
                utftext += String.fromCharCode((c >> 12) | 224);
                utftext += String.fromCharCode(((c >> 6) & 63) | 128);
                utftext += String.fromCharCode((c & 63) | 128);
            }
        }
        return utftext;
    }
    static _utf8_decode(utftext) {
        var s = "";
        var i = 0;
        var c = 0, c1 = 0, c2 = 0, c3 = 0;
        while (i < utftext.length) {
            c = utftext.charCodeAt(i);
            if (c < 128) {
                s += String.fromCharCode(c);
                i++;
            }
            else if ((c > 191) && (c < 224)) {
                c2 = utftext.charCodeAt(i + 1);
                s += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
                i += 2;
            }
            else {
                c2 = utftext.charCodeAt(i + 1);
                c3 = utftext.charCodeAt(i + 2);
                s += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
                i += 3;
            }
        }
        return s;
    }
}
Base64._keyStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
exports.Base64 = Base64;
/**
 * *****************************************************
 */
class DataUrl {
    constructor(dataUrl) {
        this._mimeType = DataUrl.getMimeType(dataUrl);
        this._data = DataUrl.getEncodedData(dataUrl);
    }
    static createDataUrl(mimeType, encodedData) {
        return DataUrl.PROTO_TOKEN + mimeType + DataUrl.ENCODING_TOKEN + encodedData;
    }
    static getMimeType(dataUrl) {
        const startIndex = dataUrl.indexOf(':');
        const endIndex = dataUrl.indexOf(';');
        if (startIndex > -1 && endIndex > startIndex) {
            return dataUrl.substring(startIndex + 1, endIndex);
        }
    }
    static getEncodedData(dataUrl) {
        const startIndex = dataUrl.indexOf(',');
        if (startIndex > -1) {
            return dataUrl.substring(startIndex + 1);
        }
    }
    get mimeType() {
        return this._mimeType;
    }
    get data() {
        return this._data;
    }
}
DataUrl.PROTO_TOKEN = 'data:';
DataUrl.ENCODING_TOKEN = ';base64,';
exports.DataUrl = DataUrl;
/**
 * *****************************************************
 */
class TimeValue {
    constructor(hours, minutes, seconds, millis) {
        this.hours = hours;
        this.minutes = minutes;
        this.seconds = seconds;
        this.millis = millis;
    }
    static fromString(timeString) {
        /* expecting hh:mm:ss.lll */
        const [hours = '0', minutes = '0', secondsPart = '0.0'] = timeString.split(':');
        const [seconds = '0', millis = '0'] = secondsPart.split('.');
        return new TimeValue(Number(hours), Number(minutes), Number(seconds), Number(millis));
    }
    static fromDateValue(dateValue) {
        return new TimeValue(dateValue.getHours(), dateValue.getMinutes(), dateValue.getSeconds(), dateValue.getMilliseconds());
    }
    toString() {
        return `${this.pad(this.hours.toString())}:${this.pad(this.minutes.toString())}:${this.pad(this.seconds.toString())}.${this.pad(this.millis.toString(), "000")}`;
    }
    toDateValue() {
        const d = new Date();
        d.setHours(this.hours, this.minutes, this.seconds, this.millis);
        return d;
    }
    pad(s, pad = "00") {
        return (pad + s).substring(s.length);
    }
}
exports.TimeValue = TimeValue;
/**
 * *****************************************************
 */
class DateValue {
    constructor(dateObj) {
        this.dateObj = dateObj;
    }
}
exports.DateValue = DateValue;
/**
 * *****************************************************
 */
class DateTimeValue {
    constructor(dateObj) {
        this.dateObj = dateObj;
    }
}
exports.DateTimeValue = DateTimeValue;
/**
 * *****************************************************
 */
var LogLevel;
(function (LogLevel) {
    LogLevel[LogLevel["ERROR"] = 0] = "ERROR";
    LogLevel[LogLevel["WARN"] = 1] = "WARN";
    LogLevel[LogLevel["INFO"] = 2] = "INFO";
    LogLevel[LogLevel["DEBUG"] = 3] = "DEBUG";
})(LogLevel = exports.LogLevel || (exports.LogLevel = {}));
class Log {
    static logLevel(level) {
        if (level >= LogLevel.DEBUG) {
            Log.debug = (message, method, clz) => {
                Log.log((o) => {
                    console.debug(o);
                }, 'DEBUG: ' + message, method, clz);
            };
        }
        else {
            Log.debug = (message, method, clz) => {
            };
        }
        if (level >= LogLevel.INFO) {
            Log.info = (message, method, clz) => {
                Log.log((o) => {
                    console.info(o);
                }, 'INFO: ' + message, method, clz);
            };
        }
        else {
            Log.info = (message, method, clz) => {
            };
        }
        if (level >= LogLevel.WARN) {
            Log.warn = (message, clz, method) => {
                Log.log((o) => {
                    console.warn(o);
                }, 'WARN: ' + message, method, clz);
            };
        }
        else {
            Log.warn = (message, clz, method) => {
            };
        }
        if (level >= LogLevel.ERROR) {
            Log.error = (message, clz, method) => {
                Log.log((o) => {
                    console.error(o);
                }, 'ERROR: ' + message, method, clz);
            };
        }
        else {
            Log.error = (message, clz, method) => {
            };
        }
        Log._logLevel = level;
    }
    static isEnabled(level) {
        return Log._logLevel >= level;
    }
    static log(logger, message, method, clz) {
        var m = typeof message !== 'string' ? Log.formatRecString(message) : message;
        if (clz || method) {
            logger(clz + "::" + method + " : " + m);
        }
        else {
            logger(m);
        }
    }
    static prettyPrint(o) {
        return ObjUtil.formatRecAttr(o, true);
    }
    static formatRecString(o) {
        return ObjUtil.formatRecAttr(o);
    }
}
//set default log level here
Log.init = Log.logLevel(LogLevel.INFO);
exports.Log = Log;
/**
 * *****************************************************
 */
class ObjUtil {
    static addAllProps(sourceObj, targetObj) {
        if (null == sourceObj || "object" != typeof sourceObj)
            return targetObj;
        if (null == targetObj || "object" != typeof targetObj)
            return targetObj;
        for (var attr in sourceObj) {
            targetObj[attr] = sourceObj[attr];
        }
        return targetObj;
    }
    static cloneOwnProps(sourceObj) {
        if (null == sourceObj || "object" != typeof sourceObj)
            return sourceObj;
        var copy = sourceObj.constructor();
        for (var attr in sourceObj) {
            if (sourceObj.hasOwnProperty(attr)) {
                copy[attr] = ObjUtil.cloneOwnProps(sourceObj[attr]);
            }
        }
        return copy;
    }
    static copyNonNullFieldsOnly(obj, newObj, filterFn) {
        for (var prop in obj) {
            if (!filterFn || filterFn(prop)) {
                var type = typeof obj[prop];
                if (type !== 'function') {
                    var val = obj[prop];
                    if (val) {
                        newObj[prop] = val;
                    }
                }
            }
        }
        return newObj;
    }
    static formatRecAttr(o, prettyPrint) {
        //@TODO - add a filter here to build a cache and detect (and skip) circular references
        if (prettyPrint) {
            return JSON.stringify(o, null, 2);
        }
        else {
            return JSON.stringify(o);
        }
    }
    static newInstance(type) {
        return new type;
    }
}
exports.ObjUtil = ObjUtil;
/**
 * *****************************************************
 */
class StringUtil {
    static splitSimpleKeyValuePair(pairString) {
        var index = pairString.indexOf(':');
        let code = null;
        let desc = '';
        if (index > -1) {
            if (index > 0)
                code = pairString.substr(0, index);
            desc = pairString.length > index ? pairString.substr(index + 1) : '';
        }
        else {
            code = pairString;
        }
        return [code, desc];
    }
    static hashCode(s) {
        var hash = 0, i, chr, len;
        if (s.length === 0)
            return hash;
        for (i = 0, len = s.length; i < len; i++) {
            chr = s.charCodeAt(i);
            hash = ((hash << 5) - hash) + chr;
            hash |= 0; // Convert to 32bit integer
        }
        return hash;
    }
    static endsWith(subjectString, searchString, position) {
        if (typeof position !== 'number' || !isFinite(position) ||
            Math.floor(position) !== position || position > subjectString.length) {
            position = subjectString.length;
        }
        position -= searchString.length;
        var lastIndex = subjectString.indexOf(searchString, position);
        return lastIndex !== -1 && lastIndex === position;
    }
}
exports.StringUtil = StringUtil;
/**
 * *****************************************************
 */
