"use strict";
/**
 * Created by rburson on 9/1/17.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const test = require("blue-tape");
const dialog_1 = require("../../catavolt/dialog");
const models_1 = require("../../catavolt/models");
const util_1 = require("../../catavolt/util");
/*
    Get a reference to the SDK instance
 */
let [tenantId, userId, password, sessionId, workbenchId, workbenchLaunchId] = ['cvtutorial', 'wsmith', 'biznes1', null, 'AAABACffAAAAAE8X', 'AAABACfaAAAAAKE8'];
let currentWorkbenches = null;
let currentRedirection = null;
let currentDialog = null;
util_1.Log.logLevel(util_1.LogLevel.DEBUG);
test("Invalid Login Test", (t) => {
    return t.shouldFail(dialog_1.Catavolt.login(tenantId, 'DESKTOP', userId, 'not a valid password')
        .catch(error => {
        //show the error we are expecting
        t.comment(`> ${JSON.stringify(error)}`);
        throw error;
    }));
});
test("Login Test", (t) => {
    return dialog_1.Catavolt.login(tenantId, 'DESKTOP', userId, password).then((session) => {
        t.ok(session, 'Expecting a non-null result');
        t.ok(session.id, 'Expecting a session id');
        t.comment(`> SessionId: ${session.id}`);
        sessionId = session.id;
        return session;
    });
});
test("Get All Workbenches Test", (t) => {
    return dialog_1.Catavolt.getWorkbenches().then((workbenches) => {
        t.ok(workbenches && workbenches.length > 0, 'Expecting a non-null result');
        currentWorkbenches = workbenches;
        workbenches.forEach((workbench, i) => {
            t.comment(`> Workbench ${i}: ${workbench.name}`);
            workbench.actions.forEach((action, i) => t.comment(`>   Action ${i}: ${action.name}`));
        });
        return workbenches;
    });
});
test("Launch Workbench By id Test", (t) => {
    return dialog_1.Catavolt.performWorkbenchActionForId(workbenchId, workbenchLaunchId)
        .then((successOrRedirection) => {
        t.ok(successOrRedirection, 'Expecting a non-null result');
        t.ok(models_1.RedirectionUtil.isDialogRedirection(successOrRedirection), 'Expecting a DialogRedirection type');
        currentRedirection = successOrRedirection;
        t.comment(`> DialogRedirection: {id: ${currentRedirection.id}, description: ${currentRedirection.dialogDescription}}`);
        return successOrRedirection;
    });
});
test("Open A DialogRedirection Test", (t) => {
    return dialog_1.Catavolt.openDialog(currentRedirection).then((dialog) => {
        t.ok(dialog, 'Expecting a non-null result');
        t.ok(dialog.id, 'Expecting a dialog id');
        t.ok(dialog.type === models_1.TypeNames.EditorDialogTypeName, 'Expecting an EditorDialog');
        t.ok(dialog.view.type === models_1.TypeNames.FormTypeName, 'Expecting a top-level Form');
        t.comment(`> Dialog: {id: ${dialog.id}, view: {type: ${dialog.view.type}}}`);
        const childObj = dialog.children.map((child) => {
            return { id: child.id, view: { type: child.view.type } };
        });
        t.comment(`>   Dialog Children: ${JSON.stringify(childObj)}`);
        currentDialog = dialog;
        return dialog;
    });
});
test("Load And Page A List Test", (t) => {
    const queryDialog = currentDialog.children[0];
    const list = queryDialog.view;
    t.comment(`List: ${queryDialog.description}`);
    t.comment(`>   List Columns: ${list.columnHeadings.join(',')}`);
    queryDialog.setScroller(5);
    return queryDialog.scroller.refresh().then((records) => {
        const rows = records.map((record) => record.propValues.join(','));
        t.comment(`>   First 5 Records:`);
        rows.forEach(row => t.comment(`>      ${row}`));
        return queryDialog.scroller.pageForward().then((records) => {
            const rows = records.map((record) => record.propValues.join(','));
            t.comment(`>   Next 5 Records:`);
            rows.forEach(row => t.comment(`>      ${row}`));
            return records;
        });
    });
});
test("Perform Action Test", (t) => {
    const queryDialog = currentDialog.children[0];
    const list = queryDialog.view;
    //Get the 'default action' on this QueryDialog
    const defaultActionId = queryDialog.defaultActionId;
    const menu = list.menu.findAtActionId(defaultActionId);
    t.ok(menu, "The View's Menu should contain the defaultActionId");
    //Choose a record to 'act on'
    const aRecord = queryDialog.scroller.buffer[2];
    t.comment(`Navigating to actionId: ${defaultActionId}`);
    //Use toDialogOrRedirection Higher-Order Function to automatically open
    //the Redirection returned by performMenuAction
    return dialog_1.Catavolt.toDialogOrRedirection(
    // Perform the action
    queryDialog.performMenuAction(menu, [aRecord.id])
        .then((successOrRedirection) => {
        if (models_1.RedirectionUtil.isDialogRedirection(successOrRedirection))
            currentRedirection = successOrRedirection;
        return successOrRedirection;
    })).then((dialogOrRedirection) => {
        if (!models_1.RedirectionUtil.isRedirection(dialogOrRedirection)) {
            const dialog = dialogOrRedirection;
            t.ok(dialog, 'Expecting a non-null result');
            t.ok(dialog.id, 'Expecting a dialog id');
            t.ok(dialog.type === models_1.TypeNames.EditorDialogTypeName, 'Expecting an EditorDialog');
            t.ok(dialog.view.type === models_1.TypeNames.FormTypeName, 'Expecting a top-level Form');
            t.comment(`> Dialog: {id: ${dialog.id}, view: {type: ${dialog.view.type}}}`);
            const childObj = dialog.children.map((child) => {
                return { id: child.id, view: { type: child.view.type } };
            });
            t.comment(`>   Dialog Children: ${JSON.stringify(childObj)}`);
            currentDialog = dialog;
        }
        return dialogOrRedirection;
    });
});
test("Read A Record From An EditorDialog Test", (t) => {
    const editorDialog = currentDialog.children[0];
    const details = editorDialog.view;
    t.comment(`Detail: ${editorDialog.description}`);
    // Get the Cell Layout defined for this Details Object
    const rowsLayout = details.rows.map((row) => {
        return row.map((cell) => {
            return cell.values.map((cellValue) => {
                return `[${cellValue.type}]`;
            }).join(', ');
        }).join(', ');
    });
    t.comment(`>  Layout is: `);
    rowsLayout.forEach(row => t.comment(`>    ${row}`));
    // Read the Record data
    //@TODO Dialog API needs a fix here for BinaryProperty
    //return editorDialog.read().then((record:Record)=>{
    //    t.ok(record);
    //    t.comment(`>  Record is: ${record.propValues.join(', ')}`);
    //   return record;
    //});
    return Promise.resolve(true);
});
test("Logout Test", (t) => {
    return dialog_1.Catavolt.logout().then((response) => {
        t.ok(response);
        t.ok(response.sessionId);
        t.equal(sessionId, response.sessionId);
        t.comment(`> Session Logout: ${response.sessionId}`);
    });
});
