/**
 * Created by rburson on 10/16/17.
 */
import { StringDictionary } from "./util";
export interface Client {
    getBlob(baseUrl: string, resourcePath?: string): Promise<BlobClientResponse>;
    getText(baseUrl: string, resourcePath?: string): Promise<TextClientResponse>;
    postMultipart<T>(baseUrl: string, resourcePath: string, formData: FormData): Promise<VoidClientResponse>;
    getJson(baseUrl: string, resourcePath: string, queryParams?: StringDictionary): Promise<JsonClientResponse>;
    postJson(baseUrl: string, resourcePath: string, body?: StringDictionary): Promise<JsonClientResponse>;
    putJson(baseUrl: string, resourcePath: string, body?: StringDictionary): Promise<JsonClientResponse>;
    deleteJson(baseUrl: string, resourcePath: string): Promise<JsonClientResponse>;
    setClientMode(clientMode: ClientMode): any;
    lastActivity: Date;
}
export declare enum ClientMode {
    ONLINE = 0,
    OFFLINE = 1,
}
export declare abstract class ClientResponse<T> {
    readonly value: T;
    readonly statusCode: number;
    constructor(value: T, statusCode: number);
}
export declare class JsonClientResponse extends ClientResponse<StringDictionary | Array<any>> {
    constructor(value: StringDictionary | Array<any>, statusCode: number);
}
export declare class BlobClientResponse extends ClientResponse<Blob> {
    constructor(value: Blob, statusCode: number);
}
export declare class TextClientResponse extends ClientResponse<string> {
    constructor(value: string, statusCode: number);
}
export declare class VoidClientResponse extends ClientResponse<void> {
    constructor(statusCode: number);
}
