import { StringDictionary } from "./util";
import { BlobClientResponse, Client, ClientMode, JsonClientResponse, TextClientResponse, VoidClientResponse } from "./client";
export declare class PersistentClient implements Client {
    private static ADD_TO_BRIEFCASE_ACTION_ID;
    private static BRIEFCASE_ACTION_ID;
    private static BRIEFCASE_DIALOG_ID;
    private static BRIEFCASE_RECORD_DIALOG_ID;
    private static BRIEFCASE_WORKPACKAGES_DIALOG_ID;
    private static ENTER_OFFLINE_MODE_ACTION_ID;
    private static EXIT_OFFLINE_MODE_ACTION_ID;
    private static OPEN_ACTION_ID;
    private static OPEN_DOCUMENT_ACTION_ID;
    private static ONLINE_PROPERTY_NAME;
    private static REFERRING_WORKBENCH_MODEL_TYPE;
    private static REFERRING_DIALOG_MODEL_TYPE;
    private static SDA_WORKBENCH_ID;
    private static WORK_PACKAGES_ACTION_ID;
    private _clientMode;
    private _fetchClient;
    private _sessionId;
    private _tenantId;
    private _userId;
    constructor(clientMode?: ClientMode);
    private _lastActivity;
    readonly lastActivity: Date;
    private writePersistentClientVars();
    private changeClientMode(clientMode);
    private changeUserId(userId);
    private changeSessionId(sessionId);
    private changeTenantId(tenantId);
    private createDialogMessageModel(message);
    private createOfflineSessionNotFoundModel();
    private createSessionIdModel(sessionId);
    private deconstructGetDialogPath(resourcePathElems);
    private deconstructGetRecordPath(resourcePathElems);
    private deconstructGetRedirectionPath(resourcePathElems);
    private deconstructPostMenuActionPath(resourcePathElems);
    private deconstructPostRecordsPath(resourcePathElems);
    private deconstructPostWorkbenchActionPath(resourcePathElems);
    deleteJson(baseUrl: string, resourcePath: string): Promise<JsonClientResponse>;
    private deleteSession(baseUrl, resourcePath);
    getBlob(baseUrl: string, resourcePath?: string): Promise<BlobClientResponse>;
    private getDialog(baseUrl, resourcePath, resourcePathElems, queryParams?);
    getJson(baseUrl: string, resourcePath?: string, queryParams?: StringDictionary): Promise<JsonClientResponse>;
    private getRecord(baseUrl, resourcePath, resourcePathElems, queryParams?);
    private getRedirection(baseUrl, resourcePath, resourcePathElems, queryParams?);
    private getSession(baseUrl, resourcePath, resourcePathElems, queryParams?);
    getText(baseUrl: string, resourcePath?: string): Promise<TextClientResponse>;
    private isWorkPackagesDialog(dialog);
    private patchWorkPackagesDialog(dialog);
    private patchWorkPackagesRecordSet(dialog, recordSet);
    private makeNullRedirectionId();
    private postAddToBriefcaseMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    postJson(baseUrl: string, resourcePath: string, jsonBody?: StringDictionary): Promise<JsonClientResponse>;
    private postMenuAction(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    postMultipart(baseUrl: string, resourcePath: string, formData: FormData): Promise<VoidClientResponse>;
    private postRecords(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    private postSession(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    private postWorkbenchAction(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    private postWorkbenchActionBriefcase(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    private postWorkbenchActionWorkPackages(baseUrl, resourcePath, resourcePathElems, jsonBody?);
    putJson(baseUrl: string, resourcePath: string, jsonBody?: StringDictionary): Promise<JsonClientResponse>;
    private readBriefcaseRecord();
    private readBriefcaseWorkpackageRecordSet();
    setClientMode(clientMode: ClientMode): void;
    private writeBriefcaseWorkpackageRecordSet(recordSet);
    private writeBriefcaseRecord(briefcaseRecord);
}
